import java.util.*;

import cj.math.jama.*;

public class DTITensor
{
    private int verbose = 0;

    private double[] b = null;

    private Matrix gradients = null;
    private Matrix Ginv = null;

    private double[] evalues = new double[3];  //max(evalues0,0)
    private double[] evalues0 = new double[3]; //original
    private Matrix evectors = new Matrix(3,3);

    private SingularValueDecomposition svd = new SingularValueDecomposition();

    private Matrix tensor = new Matrix(3,3);
    private Matrix dataM = null;

    public DTITensor () {
        // Default value for b is 700 s/mm2
        //b = new double[gradients.getRowDimension()];
        //Arrays.fill(b, 700.0);
        //System.out.println("Initializing to b-value=700");
    };

    public DTITensor(double[] b)
    {
        this.b = new double[b.length];
        System.arraycopy(b, 0, this.b, 0, b.length);
    }

    public DTITensor(Matrix gradients, double[] b)
    {
        this.gradients = gradients;
        this.b = new double[b.length];
        System.arraycopy(b, 0, this.b, 0, b.length);
        setupGinv();

        // Pre-allocate spaces.
        tensor = new Matrix(3,3);
        dataM = new Matrix(gradients.getRowDimension(), 1);
    }

    public DTITensor(Matrix gradients)
    {
        // Default value for b is 700 s/mm2
        //b = new double[gradients.getRowDimension()];
        Arrays.fill(b, 700.0);
        System.out.println("Initializing to b-value=700");
        setGradients(gradients);
    }

    public void setGradients(Matrix gradients)
    {
        this.gradients = gradients;

        setupGinv();

        // Pre-allocate spaces.
        tensor = new Matrix(3,3);
        dataM = new Matrix(gradients.getRowDimension(), 1);
    }

    /*  A function to set the list of gradients which
         will not be used in the tensor calculation.
     */
//	public void removeGradients(int[] gradient_numbers)
//	{
//		setupGinv(gradient_numbers);
//	}

    public void setBValues(double[] b)
    {
        this.b = new double[b.length];
        System.arraycopy(b, 0, this.b, 0, b.length);

        setupGinv();
    }

    final public void setVerbose(int verbose)
    {
        this.verbose = verbose;
    }

    final public double getFA()
    {
        return calcFA();
    }

    final public double getRA()
        {
            return calcRA();
        }

    final public double getVR()
        {
            return calcVR();
        }
    final public double getsRA()
        {
            return calcsRA();
        }

    final public double getinvVR()
        {
            return calcinvVR();
        }
    final public double[] getEvalues() {
        return evalues0;
    }


    final public double getADC()
    {
        return calcADC();
    }

    final public byte[] getCMAP()
    {
        byte[] cmap = new byte[3];

        cmap[0] = (byte)(Math.abs(evectors.get(0,0)) * calcFA() * 256.0);
        cmap[1] = (byte)(Math.abs(evectors.get(1,0)) * calcFA() * 256.0);
        cmap[2] = (byte)(Math.abs(evectors.get(2,0)) * calcFA() * 256.0);

        return cmap;
    }

    final public double[] getVec()
    {
        double[] vec = new double[3];

        vec[0] = evectors.get(0,0);
        vec[1] = evectors.get(1,0);
        vec[2] = evectors.get(2,0);

        return vec;
    }

    final public void setData(double[] normalized_data)
    {
        if( dataM.getRowDimension() != normalized_data.length )
        {
            dataM = new Matrix(normalized_data.length, 1);
        }

        for(int ii=0; ii<normalized_data.length; ii++)
        {
            if( !Double.isInfinite( normalized_data[ii] ) &&
                !Double.isNaN( normalized_data[ii] )  &&
                normalized_data[ii] > 0.001 )
                dataM.set(ii,0, -StrictMath.log(normalized_data[ii]));
            else
                dataM.set(ii,0, 0.0);
        }

        // Now we need to check and see if we are going
        // to remove some gradient directions.
        Matrix tensorVals = Ginv.times( dataM );

        //
        //  Setup the tensor matrix.
        //
        tensor.set(0,0, tensorVals.get(0,0));
         tensor.set(1,1, tensorVals.get(1,0));
         tensor.set(2,2, tensorVals.get(2,0));

         tensor.set(0,1, tensorVals.get(3,0));
         tensor.set(1,0, tensorVals.get(3,0));

         tensor.set(0,2, tensorVals.get(4,0));
         tensor.set(2,0, tensorVals.get(4,0));

         tensor.set(1,2, tensorVals.get(5,0));
         tensor.set(2,1, tensorVals.get(5,0));

        //
        //  Calculate the eigenvalues and eigenvectors
        //
        /*svd = tensor.svd();
        evalues = svd.getSingularValues();
        evectors = svd.getU();
        */
        EigenvalueDecomposition ed = new EigenvalueDecomposition(tensor);
        evalues0 = ed.getRealEigenvalues();
        Arrays.sort(evalues0);
        for(int i=0;i<3;i++) evalues[i]=Math.max(0,evalues0[i]);
        //evectors = ed.getV().transpose();
                svd = tensor.svd();
        //evalues = svd.getSingularValues();
        evectors = svd.getU();

    }

    final public void setData(double[] normalized_data, int[] removeGradients)
    {
        dataM = new Matrix(normalized_data.length - removeGradients.length, 1);

        for(int ii=0,jj=0; ii<normalized_data.length; ii++)
        {
            if( !isIn(ii, removeGradients) )
            {
                if( !Double.isInfinite( normalized_data[ii] ) &&
                    !Double.isNaN( normalized_data[ii] )  &&
                    normalized_data[ii] > 0.001 )
                    dataM.set(jj,0, -StrictMath.log(normalized_data[ii]));
                else
                    dataM.set(jj,0, 0.0);

                jj++;
            }
        }

        // Now we need to check and see if we are going
        // to remove some gradient directions.
        Matrix tensorVals = Ginv.times( dataM );

        //
        //  Setup the tensor matrix.
        //
        tensor.set(0,0, tensorVals.get(0,0));
        tensor.set(1,1, tensorVals.get(1,0));
        tensor.set(2,2, tensorVals.get(2,0));

        tensor.set(0,1, tensorVals.get(3,0));
        tensor.set(1,0, tensorVals.get(3,0));

        tensor.set(0,2, tensorVals.get(4,0));
        tensor.set(2,0, tensorVals.get(4,0));

        tensor.set(1,2, tensorVals.get(5,0));
        tensor.set(2,1, tensorVals.get(5,0));

        //
        //  Calculate the eigenvalues and eigenvectors
        //
        /*svd = tensor.svd();
        evalues = svd.getSingularValues();
        evectors = svd.getU();
        */
        EigenvalueDecomposition ed = new EigenvalueDecomposition(tensor);
        evalues0 = ed.getRealEigenvalues();
        Arrays.sort(evalues0);
        for(int i=0;i<3;i++) evalues[i]=Math.max(0,evalues0[i]);
        //evectors = ed.getV().transpose();
                svd = tensor.svd();
        //evalues = svd.getSingularValues();
        evectors = svd.getU();
    }

    final public Matrix getTensor()
    {
        return tensor;
    }

    private double[] calcXYZ(Matrix tensor)
    {
        Matrix V = evectors; //tensor.svd().getU();
        double[] vec = new double[3];

        for(int ii=0; ii<3; ii++) {vec[ii] = V.get(ii,0); }

        return vec;
    }

    private double calcADC()
    {
        return (evalues0[0] + evalues0[1] + evalues0[2]) / 3.0;
    }

    private double calcFA()
    {
        double sum_sq_evalues = evalues[0]*evalues[0]+
                                evalues[1]*evalues[1]+
                                evalues[2]*evalues[2];

        double sde = StrictMath.pow(evalues[0]-evalues[1], 2.0) +
                StrictMath.pow(evalues[1]-evalues[2], 2.0) +
                StrictMath.pow(evalues[0]-evalues[2], 2.0);

        double fa =  Math.sqrt( sde / (2*sum_sq_evalues) );

        return fa;
    }

    private double calcRA() //8/30/2006 BL
        {
            // see Mircostructural and Physiological Features of
            // Tissues Elucidated by Quantitative-Diffusion-Tensor MRI
            // , J. Magn. Res. 111, (1996)  equation # 14
            // RA = root(var(lambdas))/ E(lambda)
            // If I get the algebra correct
            // RA = sqrt [( lambda1 - lambda2)^2 + (lambda1 - lambda3)^2 + (lambda2-lambda3)^2 ]
            //      / (( lambda1+lambda2+lambda3)/3)

              /*
            double num = Math.sqrt((StrictMath.pow(evalues[0]-evalues[1], 2.0) +
                    StrictMath.pow(evalues[0]-evalues[2], 2.0) +
                    StrictMath.pow(evalues[1]-evalues[2], 2.0)));
            double mn_evalues = (evalues[0]+
                                 evalues[1]+
                                 evalues[2])/3;
            return ( num / mn_evalues) ;
            */
            double Dsurfsq = ((evalues[0]*evalues[1]+
                                evalues[1]*evalues[2]+
                                evalues[0]*evalues[2])/3);
            double Davsq =  StrictMath.pow((evalues[0]+
                                 evalues[1]+
                                 evalues[2])/3,2.0);

            return Math.sqrt(2*(1-Dsurfsq/Davsq));


        }


    private double calcsRA() //8/30/2006 BL
        {
            // see Mircostructural and Physiological Features of
            // Tissues Elucidated by Quantitative-Diffusion-Tensor MRI
            // , J. Magn. Res. 111, (1996)  equation # 14
            // RA = root(var(lambdas))/ E(lambda)
            // If I get the algebra correct
            // RA = sqrt [( lambda1 - lambda2)^2 + (lambda1 - lambda3)^2 + (lambda2-lambda3)^2 ]
            //      / (( lambda1+lambda2+lambda3)/3)

              /*
            double num = Math.sqrt((StrictMath.pow(evalues[0]-evalues[1], 2.0) +
                    StrictMath.pow(evalues[0]-evalues[2], 2.0) +
                    StrictMath.pow(evalues[1]-evalues[2], 2.0)));
            double mn_evalues = (evalues[0]+
                                 evalues[1]+
                                 evalues[2])/3;
            return ( num / mn_evalues) ;
            */
            double Dsurfsq = ((evalues[0]*evalues[1]+
                                evalues[1]*evalues[2]+
                                evalues[0]*evalues[2])/3);
            double Davsq =  StrictMath.pow((evalues[0]+
                                 evalues[1]+
                                 evalues[2])/3,2.0);

            return Math.sqrt((1-Dsurfsq/Davsq));


        }

    private double calcVR()
        {
            // see Towards a Quantitative Assesement of Diffusion,
            // Pierpaoli and Basser, MRM 36 (1996) p 895 equation # 2
            // VR = lambda(1)*lambda(2)*lambda(3) / [(lamda(1) + lambda(2) + lambda(3))/3]^3
             double denom =StrictMath.pow((evalues[0]+
                                 evalues[1]+
                                 evalues[2])/3,3.0);
             double num = (evalues[0]*
                                 evalues[1]*
                                 evalues[2]);

            return num/denom;
        }

    private double calcinvVR()
        {
        return 1-calcVR();
        }

    /**
     * Calculate the inverse of the gradient table
     * param gradients
     * param B value
     * return inverse of G
     */
    private void setupGinv()
    {
        Matrix G = new Matrix( gradients.getRowDimension(), 6 );

        for(int ii=0; ii<gradients.getRowDimension(); ii++)
        {
            G.set(ii,0, b[ii] * gradients.get(ii,0) * gradients.get(ii,0));
            G.set(ii,1, b[ii] * gradients.get(ii,1) * gradients.get(ii,1));
            G.set(ii,2, b[ii] * gradients.get(ii,2) * gradients.get(ii,2));
            G.set(ii,3, b[ii] * 2* gradients.get(ii,0) * gradients.get(ii,1));
            G.set(ii,4, b[ii] * 2* gradients.get(ii,0) * gradients.get(ii,2));
            G.set(ii,5, b[ii] * 2* gradients.get(ii,1) * gradients.get(ii,2));
        }

        Ginv = G.inverse();


    }

//	private void setupGinv(int[] removeGradients)
//	{
//		System.out.println("remove gradients has "+removeGradients.length+" elements");
//		System.out.println("gradients has size " + gradients.getRowDimension() + " by " + gradients.getColumnDimension());
//		Matrix G = new Matrix( gradients.getRowDimension()-removeGradients.length, 6 );
//
//		for(int ii=0,jj=0; ii<gradients.getRowDimension(); ii++)
//		{
//			if( !isIn(ii, removeGradients ) )
//			{
//				G.set(jj,0, b[ii] * gradients.get(ii,0) * gradients.get(ii,0));
//				G.set(jj,1, b[ii] * gradients.get(ii,1) * gradients.get(ii,1));
//				G.set(jj,2, b[ii] * gradients.get(ii,2) * gradients.get(ii,2));
//				G.set(jj,3, b[ii] * 2* gradients.get(ii,0) * gradients.get(ii,1));
//				G.set(jj,4, b[ii] * 2* gradients.get(ii,0) * gradients.get(ii,2));
//				G.set(jj,5, b[ii] * 2* gradients.get(ii,1) * gradients.get(ii,2));
//				jj++;
//			}
//		}
//
//		System.out.println("G is "); G.print(3,3);
//		Ginv =  G.inverse();
//	}

    final private boolean isIn(int a, int[] vec)
    {
        for(int ii=0; ii<vec.length; ii++)
            {  if( a == vec[ii]) return true; }
        return false;
    }

    public void setTensor(float v, float v1, float v2, float v3, float v4, float v5) {
            //
        //  Setup the tensor matrix.
        //
        /*if(v>0)
        System.out.println('d');
        */
        tensor.set(0,0, v);
        tensor.set(0,1, v1);
        tensor.set(0,2, v2);
        tensor.set(1,0, v1);
        tensor.set(1,1, v3);
        tensor.set(1,2, v4);
        tensor.set(2,0, v2);
        tensor.set(2,1, v4);
        tensor.set(2,2, v5);

        //
        //  Calculate the eigenvalues and eigenvectors
        //
        //System.out.println(v+" "+v1+" "+v2+" "+v3+" "+v4+" "+v5);
        /* svd = tensor.svd();
        evalues = svd.getSingularValues();
        evectors = svd.getU();        */
        EigenvalueDecomposition ed = new EigenvalueDecomposition(tensor);
        evalues0 = ed.getRealEigenvalues();
        Arrays.sort(evalues0);
        for(int i=0;i<3;i++) evalues[i]=Math.max(0,evalues0[i]);
        //evectors = ed.getV().transpose();
        svd = tensor.svd();
        //evalues = svd.getSingularValues();
        evectors = svd.getU();
    }

    public float getEval1() {return (float)evalues0[2];};
    public float getEval2() {return (float)evalues0[1];};
    public float getEval3() {return (float)evalues0[0];};

}

