import java.io.*;
import java.util.*;
import java.nio.*;
import java.nio.channels.*;
import javax.imageio.stream.*;
import javax.swing.*;
import java.awt.*;
import java.util.regex.*;

import cj.appkit.*;
import cj.math.jama.*;
import gnu.getopt.*;

public class dtiproc
{
    static public String version = "v3.5";
    static public String version_date = "April 18, 2007";

    private double sqrt3 = StrictMath.sqrt(3.0);

    //  A threshold on the signal intensity...
    //  nothing will be computed if the no diffusion image
    //  signal intensity is below this value.
    private double threshold = 10;
    private double percentile_threshold = -1;

    //  b value used...
    //private double b = 700.0;  // default //not used BL 4/17/07

    // Output directory
    String output_directory = ".";

    String[] rec_files = null;
    boolean do_scaling = true;
    ArrayList<Double> scalars = new ArrayList<Double>(10);
    String[] dpf_files = null;
    int gradients_in_rec = 0;
    String[] gradientRemovalList = null;

    private GradientsSlicesRemove gradslices_remove = new GradientsSlicesRemove();

    private DTITensor dtitensor = null;

    private DataGrad datagradients = new DataGrad();

    // Process each rec_file / dpf_file separately
    boolean process_separately = true;

    // booleans
    private boolean do_fa = false;
    private boolean do_vr = false;
    private boolean do_ra = false;
    private boolean do_ivr = false;
    private boolean do_sra = false;
    private boolean do_evals = false;
    private boolean do_cmap = false;
    private boolean do_adc = false;
    private boolean do_tensor = false;
    private boolean do_vec = false;

    private OutputFloatBuffer ofb_fa = null;
    private OutputFloatBuffer ofb_ra = null;
    private OutputFloatBuffer ofb_vr = null;
    private OutputFloatBuffer ofb_sra = null;
    private OutputFloatBuffer ofb_ivr = null;
    private OutputFloatBuffer ofb_ev1 = null;
    private OutputFloatBuffer ofb_ev2 = null;
    private OutputFloatBuffer ofb_ev3 = null;
    private OutputFloatBuffer ofb_adc = null;
    private OutputFloatBuffer ofb_dxx = null;
    private OutputFloatBuffer ofb_dyy = null;
    private OutputFloatBuffer ofb_dzz = null;
    private OutputFloatBuffer ofb_dxy = null;
    private OutputFloatBuffer ofb_dxz = null;
    private OutputFloatBuffer ofb_dyz = null;

    private String cmap_filename = "";
    private ImageOutputStream cmap_out = null;

    private String vec_filename = "";
    private ImageOutputStream ev_out = null;

    private int start_row = 0;
    private int end_row = 0;

    private int start_column = 0;
    private int end_column = 0;

    private int start_slice = 0;
    private int end_slice = 0;

    private boolean removeGradientsTF = false;
    private String removeGradientsArgument = "";

    private int verbose = 0;

    //
    //  The MAIN program.
    //
    static public void main(String[] args)
    {

        dtiproc ff = new dtiproc(args);

        System.exit(0);
    }

    private void showHelp(String filename)
    {
/*<<<<<<< .mine
		System.out.println("dtiproc [options] <REC file> <dpf file> [<REC file> <dpf file>]...");
		System.out.println("");
		System.out.println("options:");
		System.out.println("  -h | --help\t\t\tThis help");
		System.out.println("  -v \t\t\t\tIncrease the verbosity of output");
		System.out.println("  -o | --outputdir <dir>\tDefine a different output directory");
		System.out.println("  -t | --threshold <num> \tProcess data with b=0 SI greater than <num>");
		System.out.println("  --percentile-threshold <num> \tProcess data with b=0 SI greater than <num>'th percentile (e.g., 80");
		System.out.println("  --no-scaling \t\t\t\tDon't scale the input data.");
		System.out.println("  --fa \t\t\t\tOutput FA");
		System.out.println("  --adc \t\t\tOutput ADC");
		System.out.println("  --tensor \t\t\tOutput tensor");
		System.out.println("  --cmap \t\t\tOutput colormap");
		System.out.println("  --vec \t\t\tOutput eigenvector file");
		System.out.println("  --ignore [<num> | <num>:<num> ]\tRemove a gradient, or a gradient-slice");
		System.out.println("  --start-slice <num> \t\tStart processing on slice <num>");
		System.out.println("  --end-slice <num> \t\tStop processing on slice <num>");
		System.out.println("  -p \t\t\t\tProcess multiple REC files together.");
		System.out.println("");
		System.out.println("Notes:");
		System.out.println(" - By default, the REC files will be processed separately, as if the program was run");
		System.out.println("   once for each REC file.");
		System.out.println(" - If the -p option is used, the SI vector and gradient matrix will be N times longer");
		System.out.println("   where N is the number of REC file/dpf file pairs");
		System.out.println(" - Should always use some threshold value to get around low signal data outside the head.");
		System.out.println(" - All data is output in little endian floating point data except tensor and color map.");
		System.out.println(" - The FA and VEC output may be loaded directly into DTIStudio to create fiber tracks.");
		System.out.println("");
		System.out.println("   Written by Craig Jones (craig@mri.jhu.edu).");
		System.out.println("   Use at your own risk!!");
=======*/
        BufferedReader in = null;
        String line = "";
//>>>>>>> .r13

        try {
            InputStream is = getClass().getResourceAsStream(filename);
            BufferedReader br = new BufferedReader(new InputStreamReader(is));

            while( (line = br.readLine()) != null )
                System.out.println(line);
        }
        catch(IOException e )
        {
            System.out.println(e);
            System.exit(-1);
        }
    }

    private void processOptions(String[] argv)
    {
        int c;
        String arg;
        int help = 0;
        StringBuffer temp = new StringBuffer();
/*<<<<<<< .mine
		 LongOpt[] longopts = new LongOpt[13];
=======
*/		 LongOpt[] longopts = new LongOpt[22];
//>>>>>>> .r13
        //
        StringBuffer sb_output_directory = new StringBuffer(output_directory);
        longopts[0] = new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h');
        longopts[1] = new LongOpt("outputdir",
                LongOpt.REQUIRED_ARGUMENT, sb_output_directory, 'o');

        // FA
        longopts[2] = new LongOpt("fa", LongOpt.NO_ARGUMENT, null, 2);

        // RA
        longopts[18] = new LongOpt("ra", LongOpt.NO_ARGUMENT, null, 18);

        // VR
        longopts[19] = new LongOpt("vr", LongOpt.NO_ARGUMENT, null, 19);

        // sRA
        longopts[20] = new LongOpt("sra", LongOpt.NO_ARGUMENT, null, 20);

        // 1-VR
        longopts[21] = new LongOpt("invvr", LongOpt.NO_ARGUMENT, null, 21);

        // ADC
        longopts[3] = new LongOpt("adc", LongOpt.NO_ARGUMENT, null, 3);

        // Tensor
        longopts[4] = new LongOpt("tensor", LongOpt.NO_ARGUMENT, null, 4);

        // CMAP
        longopts[5] = new LongOpt("cmap", LongOpt.NO_ARGUMENT, null, 5);

        // Start slice
        longopts[6] = new LongOpt("start-slice", LongOpt.REQUIRED_ARGUMENT, null, 6);

        // End Slice
        longopts[7] = new LongOpt("end-slice", LongOpt.REQUIRED_ARGUMENT, null, 7);

        // vec
        longopts[8] = new LongOpt("vec", LongOpt.NO_ARGUMENT, null, 8);

        // evalues
        longopts[17] = new LongOpt("evalues", LongOpt.NO_ARGUMENT, null, 17);

        // threshold
        longopts[9] = new LongOpt("threshold", LongOpt.REQUIRED_ARGUMENT, null, 't');

        // no-scaling
        longopts[10] = new LongOpt("no-scaling", LongOpt.NO_ARGUMENT, null, 10);

        // percentile-threshold
        longopts[11] = new LongOpt("percentile-threshold", LongOpt.REQUIRED_ARGUMENT, null, 11);

/*<<<<<<< .mine
		 // ignore
		 longopts[12] = new LongOpt("ignore", LongOpt.REQUIRED_ARGUMENT, null, 12); 

=======
*/		 // percentile-threshold
        longopts[12] = new LongOpt("remove-gradient-directions", LongOpt.REQUIRED_ARGUMENT, null, 12);

        // start-row
        longopts[13] = new LongOpt("start-row", LongOpt.REQUIRED_ARGUMENT, null, 13);

        // end-row
        longopts[14] = new LongOpt("end-row", LongOpt.REQUIRED_ARGUMENT, null, 14);

        // start-column
        longopts[15] = new LongOpt("start-column", LongOpt.REQUIRED_ARGUMENT, null, 15);

        // end-column
        longopts[16] = new LongOpt("end-column", LongOpt.REQUIRED_ARGUMENT, null, 16);

//>>>>>>> .r13
        //
        Getopt g = new Getopt("dtiproc", argv, "vhs:t:o:pW;", longopts);

        while ((c = g.getopt()) != -1)
            switch (c)
            {
                case 0:
                    arg = g.getOptarg();
//				  System.out.println("Got long option with value '" +
//									 (char)(new Integer(sb_output_directory.toString())).intValue()
//									 + "' with argument " +
//									 ((arg != null) ? arg : "null"));
                    break;
                    //
                case 1:
//				  System.out.println("I see you have return in order set and that " +
//									 "a non-option argv element was just found " +
//									 "with the value '" + g.getOptarg() + "'");
                    break;
                    //  FA
                case 2:
                    do_fa = true;
                    break;

                    //  RA
                case 18:
                    do_ra = true;
                    break;

                    //  VR
                case 19:
                    do_vr = true;
                    break;

                    //  sRA
                case 20:
                    do_sra = true;
                    break;

                    //  VA
                case 21:
                    do_ivr = true;
                    break;

                case 17:
                    do_evals = true;
                    break;

                    //
                case 3:
                    do_adc = true;
                    break;
                    //
                case 4:
                    do_tensor = true;
                    break;
                    //
                case 5:
                    do_cmap = true;
                    break;
                    //
                case 6:
                    try {
                        start_slice = Integer.parseInt( g.getOptarg() );
                    }
                    catch( NumberFormatException e )
                    {
                        System.out.println("For start-slice " + g.getOptarg() + " is not a number");
                    }
                    break;
                    //
                case 7:
                    try {
                        end_slice = Integer.parseInt( g.getOptarg() );
                    }
                    catch( NumberFormatException e )
                    {
                        System.out.println("For end-slice " + g.getOptarg() + " is not a number");
                    }
                    break;
                    //
                case 8:
                    do_vec = true;
                    break;
                    //
                case 'v':
                    verbose++;
                    break;
                    //
                case 'o':
                    sb_output_directory = new StringBuffer(g.getOptarg());
                    break;
                    //
                case 10:
                    do_scaling = false;
                    break;
                case 11:
                    try {
                        percentile_threshold = Double.parseDouble( g.getOptarg() );
                    }
                    catch( NumberFormatException e )
                    {
                        System.out.println("percentile-threshold " + g.getOptarg() + " is not a number");
                        System.exit(0);
                    }

                    if( percentile_threshold < 0.0 || percentile_threshold > 100.0 )
                    {
                        System.out.println("percentile-threshold " + g.getOptarg() + " must be between 0 and 100 (%)");
                        System.exit(0);

                    }
                    break;
                case 12:  // Include only specific gradient directions.
                    removeGradientsArgument = g.getOptarg();
                    removeGradientsTF = true;
                    break;
                    //
                    /*case 12:
                       gradslices_remove.remove( g.getOptarg());
                       break; */
                    //
                case 't':
                    threshold = (new Double(g.getOptarg())).doubleValue();
                    System.out.println("threshold is " + threshold);
                    break;
                    //
                    //
                case 13:
                    System.out.println("start-row not implemented correctly, do not use yet.");
                    System.exit(0);
                    try {
                        start_row = Integer.parseInt( g.getOptarg() );
                    }
                    catch( NumberFormatException e )
                    {
                        System.out.println("For start-row " + g.getOptarg() + " is not a number");
                    }
                    break;
                    //
                case 14:
                    System.out.println("end-row not implemented correctly, do not use yet.");
                    System.exit(0);
                    try {
                        end_row = Integer.parseInt( g.getOptarg() );
                    }
                    catch( NumberFormatException e )
                    {
                        System.out.println("For end-row " + g.getOptarg() + " is not a number");
                    }
                    break;
                    //
                case 15:
                    System.out.println("start-column not implemented correctly, do not use yet.");
                    System.exit(0);
                    try {
                        start_column = Integer.parseInt( g.getOptarg() );
                    }
                    catch( NumberFormatException e )
                    {
                        System.out.println("For start-column " + g.getOptarg() + " is not a number");
                    }
                    break;
                    //
                case 16:
                    System.out.println("end-column not implemented correctly, do not use yet.");
                    System.exit(0);
                    try {
                        end_column = Integer.parseInt( g.getOptarg() );
                    }
                    catch( NumberFormatException e )
                    {
                        System.out.println("For end-column " + g.getOptarg() + " is not a number");
                    }
                    break;
                case 'h':
                    help++;
                    break;
                    //
                case 'p':
                    process_separately = false;
                    break;
                    //
                case 'W':
//				  System.out.println("Hmmm. You tried a -W with an incorrect long " +
//									 "option name");
                    break;
                    //
                case ':':
                    System.out.println("Doh! You need an argument for option " +
                            (char)g.getOptopt());
                    break;
                    //
                case '?':
                    System.out.println("The option '" + (char)g.getOptopt() +
                            "' is not valid");
                    break;
                    //
                default:
                    break;
            }

        if( help > 0 )
            showHelp("help.txt");

        if( help > 1 )
            showHelp("help_notes.txt");

        //
        rec_files = new String[ (argv.length - g.getOptind())/2 ];
        dpf_files =new String[ (argv.length - g.getOptind())/2 ];
        gradientRemovalList =new String[ (argv.length - g.getOptind())/2 ];

        Pattern comma_pat = Pattern.compile("[\\d,]+");
        Matcher comma_mat = null;

        for (int i = g.getOptind(), jj=0; i < argv.length ; i+=2,jj++)
        {
            // Grab the rec file and DPF file
            rec_files[jj] = argv[i];
            dpf_files[jj] = argv[i+1];

            // If there are more arguments, then check
            // to see if the next argument is the gradient list
            // of which to exclude.
            if( i+2 < argv.length )
            {
                comma_mat = comma_pat.matcher(argv[i+2]);
                if( comma_mat.matches() )
                {
                    gradientRemovalList[jj] = argv[i+2];
                }
            }
            else
                comma_mat = comma_pat.matcher("a");

            // Be verbose about the output if we want
            // to see something.
            if( verbose > 0 )
            {
                System.out.print("rec/dpf = " +
                        rec_files[jj] + " " + dpf_files[jj]);

                if( comma_mat.matches() )
                    System.out.println("  " + argv[i+2] + " is the set of dwi/slices to restrict.");
                else
                    System.out.println("");

            }

            if( comma_mat.matches() )
            {
                i++;
            }
        }

        /*
           *  Make sure the scalars are setup.
           */
        if( scalars.size() > 0 )
        {
            if( scalars.size() != rec_files.length )
            {
                System.out.println("dtiproc: There are " + scalars.size() + " scaling factors, but " + rec_files.length + " REC files - the number must match.");
            }
        }
        else
        {
            for(int ii=0; ii< rec_files.length; ii++)
                scalars.add( 1.0 );
        }

        output_directory = sb_output_directory.toString();

        if( verbose > 0 )
        {
            System.out.println("output directory = " + output_directory);
        }
    }

    public dtiproc(String[] args)
    {
        System.out.println("-----------------------------------------------");
        System.out.println("dtiproc " + version + " (" + version_date + ")");
        System.out.println("Written by Craig Jones (craig@mri.jhu.edu)");
        System.out.println("Updated by Bennett Landman (bennett@bme.jhu.edu)");
        System.out.println("-----------------------------------------------");

        processOptions(args);

        if( process_separately )
            processEach();
        else
            processTogether();
    }

    /**
     * Open all the output files which were selected on the command line
     * @param rec_filename
     * @returns nothing
     */
    private void openOutputFiles(String rec_filename, int nrows, int ncols, int nslices)
    {
        //
        //   FA
        //
        if( do_fa )
        {
            ofb_fa =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "fa");
        }
        if( do_ra )
        {
            ofb_ra =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "ra");
        }
        if( do_vr )
        {
            ofb_vr =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "vr");
        }
        if( do_sra )
        {
            ofb_sra =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "sra");
        }
        if( do_ivr )
        {
            ofb_ivr =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "invvr");
        }

        if( do_evals ){
            ofb_ev1 =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "eval1");
            ofb_ev2 =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "eval2");
            ofb_ev3 =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "eval3");

        }

        //
        //  Tensor
        //
        if( do_tensor )
        {
            ofb_dxz =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "dxz");

            ofb_dxy =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "dxy");

            ofb_dyz =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "dyz");

            ofb_dxx =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "dxx");

            ofb_dyy =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "dyy");

            ofb_dzz =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "dzz");
        }

        //
        //  ADC
        //
        if( do_adc )
        {
            ofb_adc =
                    new OutputFloatBuffer(rec_filename, output_directory, nrows, ncols, nslices, "adc");
        }

        //
        //  CMAP.
        //
        if( do_cmap )
        {
            String temp = (new File( rec_filename )).getName();
            cmap_filename = output_directory + File.separator +
                    temp.substring(0, temp.length()-4) + ".cmap";

            cmap_out = AppKit.openWrite( cmap_filename );
            cmap_out.setByteOrder( ByteOrder.LITTLE_ENDIAN );
        }

        //
        //  Vec
        //
        if( do_vec )
        {
            String temp = (new File( rec_filename )).getName();
            vec_filename = output_directory + File.separator +
                    temp.substring(0,temp.length()-4) + ".vec";

            ev_out = AppKit.openWrite( vec_filename );
            ev_out.setByteOrder( ByteOrder.LITTLE_ENDIAN );
        }
    }

    public void processEach()
    {
        for(int filesi=0; filesi < rec_files.length; filesi++)
        {
            String rec_filename = rec_files[filesi];
            String dpf_filename = dpf_files[filesi];
            String gradientRemoval = gradientRemovalList[filesi];

            if( removeGradientsTF )
                gradientRemoval = gradientRemoval +", " + removeGradientsArgument;

            //--------------------------------------------------------------------
            //
            //  Setup the gradients.
            //
            //--------------------------------------------------------------------
            Gradients gradients_obj = new Gradients(dpf_filename);

            //--------------------------------------------------------------------
            //
            //  Setup the PARREC
            //
            //--------------------------------------------------------------------
            PARREC parrec = new PARREC(rec_filename, do_scaling);

            //--------------------------------------------------------------------
            //
            //  Setup the DTITensor processsor.
            //
            //--------------------------------------------------------------------
            dtitensor = new DTITensor(parrec.getBValues());

            if( end_row == 0 ) { end_row = parrec.getNRows(); }
            if( end_column == 0 ) { end_column = parrec.getNCols(); }
            if( end_slice == 0 ) { end_slice = parrec.getNSlices(); }

            openOutputFiles(rec_filename, parrec.getNRows(), parrec.getNCols(), parrec.getNSlices());

            //----------------------------------------------------------------
            //
            //   Setup the DataGradients container
            //
            //----------------------------------------------------------------
            datagradients = new DataGrad( );

            if( gradientRemoval == null )
                datagradients.add( parrec, gradients_obj );
            else
                datagradients.add( parrec, gradients_obj, gradientRemoval );

            //----------------------------------------------------------------
            //
            //   Print out interesting information
            //
            //----------------------------------------------------------------
            if( verbose > 0 )
            {
                System.out.println("Information:");
                System.out.println("Data file " + rec_filename);
                System.out.println("Gradient file " + dpf_filename);
                if( do_fa )
                {
                    System.out.println("FA filename " +
                            ofb_fa.getFilename());
                }
                if( do_ra )
                {
                    System.out.println("RA filename " +
                            ofb_ra.getFilename());
                }
                if( do_vr )
                {
                    System.out.println("VR filename " +
                            ofb_vr.getFilename());
                }
                if( do_sra )
                {
                    System.out.println("sRA filename " +
                            ofb_sra.getFilename());
                }
                if( do_ivr )
                {
                    System.out.println("invVR filename " +
                            ofb_vr.getFilename());
                }
                if( do_evals )
                {
                    System.out.println("Eval1 filename " +
                            ofb_ev1.getFilename());
                    System.out.println("Eval2 filename " +
                            ofb_ev2.getFilename());
                    System.out.println("Eval3 filename " +
                            ofb_ev3.getFilename());
                }

                if( do_adc )
                {
                    System.out.println("ADC filename " +
                            ofb_adc.getFilename());
                }

                if( do_tensor )
                {
                    System.out.println("Dxx filename " +
                            ofb_dxx.getFilename());
                    System.out.println("Dyy filename " +
                            ofb_dyy.getFilename());
                    System.out.println("Dzz filename " +
                            ofb_dzz.getFilename());
                    System.out.println("Dxy filename " +
                            ofb_dxy.getFilename());
                    System.out.println("Dxz filename " +
                            ofb_dxz.getFilename());
                    System.out.println("Dyz filename " +
                            ofb_dyz.getFilename());
                }

                if( do_cmap )
                {
                    System.out.println("Colormap filename " +
                            cmap_filename);
                }

                if( do_vec )
                    System.out.println("Eigenvector filename " + vec_filename);

                System.out.println("Rows, columns and slices = " +
                        parrec.getNRows() + ", " + parrec.getNCols() + ", " + parrec.getNSlices());

                Matrix temp = datagradients.getGradients(0);
                System.out.println("The gradient matrix (of slice 0) is of size " + temp.getRowDimension() + " by " + temp.getColumnDimension() );

                if( verbose > 1 )
                    temp.print(3,3);
            }

            //----------------------------------------------------------------
            //
            //   Create the Progress Bar to show how fast we are
            //   reading things in...
            //
            //----------------------------------------------------------------
            CJProgressBar cjpb = new CJProgressBar("Processing " + (new File(rec_filename)).getName() + "...", parrec.getNSlices());

            long time_start = System.currentTimeMillis();
            int time_counter = 0;
            int total_counts = parrec.getNSlices();

            //
            //  Determine threshold
            //
            if( percentile_threshold > 0.0 )
            {
                threshold = parrec.getPercentileThreshold( percentile_threshold );
                System.out.println("percentile threshold is " + percentile_threshold + " which results in a SI threshold of " + threshold);
            }

            //----------------------------------------------------------------
            //
            // Run through the file.
            //
            //   I had some problems with blocking, likely it was trying to
            //   do the SVD with some values inf/nan because of bad log(0)
            //
            //----------------------------------------------------------------
            for(int slice=start_slice; slice<end_slice;slice++){

                dtitensor.setGradients( datagradients.getGradients(slice) );

                cjpb.update( slice );

                for(int row=start_row; row<end_row;row++){
                    for(int col=start_column; col<end_column;col++){

                        double ref = datagradients.getReference(row,col,slice);
                        double[] all_data = datagradients.getData(row,col,slice);

                        processNormalizedData(all_data, ref > threshold, row, col, slice);
                    }
                }

            }

            cjpb.close();

            //
            //  We are done writing out, so close all the files.
            //
            try{
                if( do_vec) { ev_out.close(); }

                if( do_fa ) { ofb_fa.close(); }
                if( do_ra ) { ofb_ra.close(); }
                if( do_vr ) { ofb_vr.close(); }
                if( do_sra ) { ofb_sra.close(); }
                if( do_ivr ) { ofb_ivr.close(); }
                if( do_evals ) { ofb_ev1.close(); ofb_ev2.close(); ofb_ev3.close();}
                if( do_cmap ) { cmap_out.close(); }
                if( do_adc ) ofb_adc.close();

                if( do_tensor ) {
                    ofb_dxx.close(); ofb_dyy.close(); ofb_dzz.close();
                    ofb_dxy.close(); ofb_dxz.close(); ofb_dyz.close();
                }
            } catch(IOException e) {System.out.println(e); System.exit(-1);}

        }
    }

    public void processTogether()
    {
        //--------------------------------------------------------------------
        //
        //  Create the array of PARREC objects for each REC file.
        //
        //--------------------------------------------------------------------
        for(int ii=0; ii < rec_files.length; ii++)
        {
            if( removeGradientsTF )
            {
                datagradients.add( new PARREC(rec_files[ii], do_scaling),
                        new Gradients( dpf_files[ii] ),
                        gradientRemovalList[ii] );
            }
            else
            {
                datagradients.add( new PARREC(rec_files[ii], do_scaling),
                        new Gradients( dpf_files[ii] ),
                        null);
                //gradientRemovalList[ii] + "," + removeGradientsArgument );
            }


        }


        openOutputFiles(rec_files[0], datagradients.getNRows(),
                datagradients.getNCols(),
                datagradients.getNSlices());

        //--------------------------------------------------------------------
        //
        //  Setup the large dtitensor.
        //
        //--------------------------------------------------------------------
        double[] btemp = datagradients.getBValues();
        dtitensor = new DTITensor(btemp);

        if( end_row == 0 ) { end_row = datagradients.getNRows(); }
        if( end_column == 0 ) { end_column = datagradients.getNCols(); }
        if( end_slice == 0 ) { end_slice = datagradients.getNSlices(); }

        //----------------------------------------------------------------
        //
        //   Print out interesting information
        //
        //----------------------------------------------------------------
        if( verbose > 0 )
        {
            System.out.println("Information:");
            System.out.println("Data file " + rec_files[0]);
            System.out.println("Gradient file " + dpf_files[0]);
            if( do_fa )
            {
                System.out.println("FA filename " +
                        ofb_fa.getFilename());
            }
            if( do_ra )
            {
                System.out.println("RA filename " +
                        ofb_ra.getFilename());
            }
            if( do_vr )
            {
                System.out.println("VR filename " +
                        ofb_vr.getFilename());
            }
            if( do_sra )
            {
                System.out.println("sRA filename " +
                        ofb_sra.getFilename());
            }
            if( do_ivr )
            {
                System.out.println("invVR filename " +
                        ofb_ivr.getFilename());
            }
            if( do_evals )
            {
                System.out.println("Eval1 filename " +
                        ofb_ev1.getFilename());
                System.out.println("Eval2 filename " +
                        ofb_ev2.getFilename());
                System.out.println("Eval3 filename " +
                        ofb_ev3.getFilename());
            }

            if( do_adc )
            {
                System.out.println("ADC filename " +
                        ofb_adc.getFilename());
            }

            if( do_tensor )
            {
                System.out.println("Dxx filename " +
                        ofb_dxx.getFilename());
                System.out.println("Dyy filename " +
                        ofb_dyy.getFilename());
                System.out.println("Dzz filename " +
                        ofb_dzz.getFilename());
                System.out.println("Dxy filename " +
                        ofb_dxy.getFilename());
                System.out.println("Dxz filename " +
                        ofb_dxz.getFilename());
                System.out.println("Dyz filename " +
                        ofb_dyz.getFilename());
            }

            if( do_cmap )
            {
                System.out.println("Colormap filename " +
                        cmap_filename);
            }

            if( do_vec )
                System.out.println("Eigenvector filename " + vec_filename);

            System.out.println("Rows, columns and slices = " +
                    datagradients.getNRows() + ", " + datagradients.getNCols() + ", " + datagradients.getNSlices());

            Matrix temp = datagradients.getGradients(0);
            System.out.println("The gradient matrix (of slice 0) is of size " + temp.getRowDimension() + " by " + temp.getColumnDimension() );

            if( verbose > 1 )
                temp.print(3,3);
        }

        //----------------------------------------------------------------
        //
        //   Create the Progress Bar to show how fast we are
        //   reading things in...
        //
        //----------------------------------------------------------------
        CJProgressBar cjpb = new CJProgressBar("Processing " + (new File(rec_files[0])).getName() + "...", datagradients.getNSlices());

        //
        //  Determine threshold
        //
        if( percentile_threshold > 0.0 )
        {
            threshold = datagradients.getPercentileThreshold( percentile_threshold );
            System.out.println("percentile threshold is " + percentile_threshold + " which results in a SI threshold of " + threshold);
        }

        //----------------------------------------------------------------
        //
        // Run through the file.
        //
        //   I had some problems with blocking, likely it was trying to
        //   do the SVD with some values inf/nan because of bad log(0)
        //
        //----------------------------------------------------------------
        for(int slice=start_slice; slice<end_slice;slice++){

            /*
                *  Now setup the gradients which might depend on
                *  the slice.
                */
            dtitensor.setGradients(datagradients.getGradients(slice) );

            cjpb.update( slice );
            for(int row=start_row; row<end_row;row++){
                for(int col=start_column; col<end_column;col++){

                    double ref = datagradients.getReference(row,col,slice);
                    double[] all_data = datagradients.getData(row,col,slice);

                    processNormalizedData(all_data, ref > threshold, row, col, slice);
                }
            }

        }

        cjpb.close();

        //
        //  We are done writing out, so close all the files.
        //
        try{
            if( do_vec) { ev_out.close(); }

            if( do_fa ) { ofb_fa.close(); }
            if( do_ra ) { ofb_ra.close(); }
            if( do_vr ) { ofb_vr.close(); }
            if( do_sra ) { ofb_sra.close(); }
            if( do_ivr ) { ofb_ivr.close(); }
            if( do_evals ) { ofb_ev1.close(); ofb_ev2.close(); ofb_ev3.close();}
            if( do_cmap ) { cmap_out.close(); }
            if( do_adc ) ofb_adc.close();

            if( do_tensor ) {
                ofb_dxx.close(); ofb_dyy.close(); ofb_dzz.close();
                ofb_dxy.close(); ofb_dxz.close(); ofb_dyz.close();
            }
        } catch(IOException e) {System.out.println(e); System.exit(-1);}

    }

    final private void processNormalizedData(double[] normalized_data, boolean doit, int row, int col, int slice)
    {

        if( verbose > 3 )
            System.out.println("In processNormalizedData");

        if( doit )
        {
            if( verbose > 3 )
                System.out.println("Going to set the normalized vector");

            // Very important command, this is where everything happens!!!!!
            dtitensor.setData(normalized_data);
            if(do_evals) {
                double ee[] = dtitensor.getEvalues();
                if( Double.isInfinite(ee[0]) || Double.isNaN(ee[0]) )
                {
                    ofb_ev1.put((float)0.0);
                    ofb_ev2.put((float)0.0);
                    ofb_ev3.put((float)0.0);
                }
                else
                    ofb_ev1.put((float)ee[0]);
                ofb_ev2.put((float)ee[1]);
                ofb_ev3.put((float)ee[2]);
            }
            if( do_fa )
            {
                double ff = dtitensor.getFA();
                if( Double.isInfinite(ff) || Double.isNaN(ff) )
                {
                    ofb_fa.put((float)0.0);
                }
                else
                    ofb_fa.put((float)ff);
            };
            if( do_ra )
            {
                double ff = dtitensor.getRA();
                if( Double.isInfinite(ff) || Double.isNaN(ff) )
                {
                    ofb_ra.put((float)0.0);
                }
                else
                    ofb_ra.put((float)ff);
            };
            if( do_vr )
            {
                double ff = dtitensor.getVR();
                if( Double.isInfinite(ff) || Double.isNaN(ff) )
                {
                    ofb_vr.put((float)0.0);
                }
                else
                    ofb_vr.put((float)ff);
            };
            if( do_sra )
            {
                double ff = dtitensor.getsRA();
                if( Double.isInfinite(ff) || Double.isNaN(ff) )
                {
                    ofb_sra.put((float)0.0);
                }
                else
                    ofb_sra.put((float)ff);
            };
            if( do_ivr )
            {
                double ff = dtitensor.getinvVR();
                if( Double.isInfinite(ff) || Double.isNaN(ff) )
                {
                    ofb_ivr.put((float)0.0);
                }
                else
                    ofb_ivr.put((float)ff);
            };

            if( do_tensor ) {
                Matrix tensorVals = dtitensor.getTensor();

                ofb_dxx.put((float)tensorVals.get(0,0));
                ofb_dyy.put((float)tensorVals.get(1,1));
                ofb_dzz.put((float)tensorVals.get(2,2));

                ofb_dxy.put((float)tensorVals.get(0,1));
                ofb_dxz.put((float)tensorVals.get(0,2));
                ofb_dyz.put((float)tensorVals.get(1,2));
            }

            if( do_adc ) ofb_adc.put((float)dtitensor.getADC());

            if( do_vec )
            {
                try {
                    double[] vec = dtitensor.getVec();

                    if( Double.isInfinite(vec[0]) || Double.isNaN(vec[0]) ||
                            Double.isInfinite(vec[1]) || Double.isNaN(vec[1]) ||
                            Double.isInfinite(vec[2]) || Double.isNaN(vec[2]) )
                    {
                        System.out.println("Problem with VEC value.");
                        System.exit(-1);
                    }
                    else
                    {
                        ev_out.writeFloat((float)vec[0]);
                        ev_out.writeFloat((float)vec[1]);
                        ev_out.writeFloat((float)vec[2]);
                    }
                }
                catch( Exception e ) { System.out.println(e); System.exit(0); }
            }

            if( do_cmap )
            {
                try {
                    byte[] cm = dtitensor.getCMAP();
                    cmap_out.writeByte(cm[0]);
                    cmap_out.writeByte(cm[1]);
                    cmap_out.writeByte(cm[2]);
                }
                catch( Exception e ) { System.out.println(e); System.exit(0); }
            }
        }
        else   // If the threshold isn't met, write out zeros
        {
            if( do_fa ) ofb_fa.put((float)0.0);
            if( do_ra ) ofb_ra.put((float)0.0);
            if( do_vr ) ofb_vr.put((float)0.0);
            if( do_sra ) ofb_sra.put((float)0.0);
            if( do_ivr ) ofb_ivr.put((float)0.0);
            if( do_adc ) ofb_adc.put((float)0.0);

            if(do_evals) {

                ofb_ev1.put((float)0.0);
                ofb_ev2.put((float)0.0);
                ofb_ev3.put((float)0.0);

            }
            if( do_vec )
            {
                try {
                    ev_out.writeFloat((float)0.0);
                    ev_out.writeFloat((float)0.0);
                    ev_out.writeFloat((float)0.0);
                }
                catch( Exception e ) { System.out.println(e); System.exit(0); }
            }

            if( do_tensor ) {
                ofb_dxx.put((float)0.0);
                ofb_dyy.put((float)0.0);
                ofb_dzz.put((float)0.0);

                ofb_dxy.put((float)0.0);
                ofb_dxz.put((float)0.0);
                ofb_dyz.put((float)0.0);
            }

            if( do_cmap )
            {
                try {
                    cmap_out.writeByte((int)0);
                    cmap_out.writeByte((int)0);
                    cmap_out.writeByte((int)0);
                }
                catch( Exception e ) { System.out.println(e); System.exit(0); }
            }
        }

    }

    //
    //  If we want to set the threshold
    //
    public void setThreshold(double new_thresh)
    {  threshold = new_thresh; }

}

//  This is an output buffer which we will
//  use to write out the data.
class OutputFloatBuffer
{
    FileOutputStream fos = null;
//	BufferedOutputStream bos = null;
    FileChannel fc = null;
    ByteBuffer bb = null;
    String output_filename;


    public OutputFloatBuffer(String rec_filename, String output_dir,
                             int nrows, int ncols, int nslices,
                             String ext)
    {
        open(rec_filename, output_dir, nrows, ncols, nslices, ext);
    }


    private void open(String rec_filename, String output_dir,
                      int nrows, int ncols, int nslices,
                      String ext)
    {
        String temp = (new File( rec_filename )).getName();
        output_filename = output_dir + File.separator + temp.substring(0, temp.length()-4) + "." + ext;

        try {
            fos = new FileOutputStream(new File(output_filename));
            fc = fos.getChannel();

            bb = ByteBuffer.allocateDirect(nrows*ncols*nslices*4);
            bb.order( ByteOrder.LITTLE_ENDIAN);
        }
        catch( IOException e )
        {
            System.out.println(e);
            System.exit(-1);
        }

    }

    public void close()
    {
        try {
            bb.flip();
            fc.write(bb);
            //bb.close();
            fc.close();
        }
        catch( IOException e )
        {
            System.out.println("Could not close " + output_filename);
            System.out.println(e);
        }
    }

    final public void put(float val)
    { bb.putFloat(val); }

    public String getFilename()
    { return output_filename; }
}


class CJProgressBar
{
    private boolean text = false;
    private long time_start = System.currentTimeMillis();
    private int numItems = 0;
    JFrame jf = null;
    JProgressBar progressBar = null;

    public CJProgressBar(String title, int numItems)
    {
        this.numItems = numItems;

        time_start = System.currentTimeMillis();

        try {
            jf = new JFrame(title);
            jf.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );

            progressBar = new JProgressBar(0, numItems);
            progressBar.setValue(0);
            progressBar.setMinimum(0);
            progressBar.setMaximum(numItems);
            progressBar.setStringPainted(true);
            progressBar.setFont(new Font(null, Font.PLAIN, 10));

            jf.getContentPane().add(progressBar);
            Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
            jf.setSize(200,60);
            jf.setLocation(screen.width/2-200/2,screen.height/2-30/2);
            jf.setVisible(true);
        }
        catch( Exception e )
        {
            text = true;
        }

        if( text )
        {

        }
    }

    public void update(int iter)
    {
        //  Update the progress.
        long time_current = System.currentTimeMillis();

        int time_left = (int) ( (time_current-time_start)/(float)(iter+1) *
                (float)(numItems - iter) / (float)1000.0);

        if( !text )
        {
            progressBar.setValue( iter );
            progressBar.setString("Approx time remaining " + time_left + " seconds");
        }
        else
        {
            System.out.print("\r " + iter + "/" + numItems + ": Approx time remaining " + time_left + " seconds");
        }
    }

    public void close()
    {
        if( !text )
        {
            jf.setVisible(false);
        }
        else
        {
            System.out.println("Done");
        }
    }
}
