package apps.affine;

import cj.fileformats.*;
import cj.volume.MCVolume;
import cj.volume.process.Affine;
import cj.appkit.FilterAppKit;

final public class AffineApp {
   public static void main(String argv[]) {
      AffineKit myApp = new AffineKit(); myApp.main(argv);
   }
}

final class AffineKit extends FilterAppKit {

    private String comment = "";
	private String output_suffix = "_affine";
	private Affine af = null;
	private double trans_x = 0.0;
	private double trans_y = 0.0;
	private double trans_z = 0.0;

	public String getAppName() { return "affine"; }
	public String getDialogName() { return "Process image"; }
    public String getOutputSuffix() { return output_suffix; }
    public String getComment() { return comment; }

    protected void moreOpts() { 
        addOpt_reqArg("output-suffix", 800, "Output suffix (def: 'A').");
        addOpt_reqArg("translate-x", 801, "Translate X (columns).");
        addOpt_reqArg("translate-y", 802, "Translate Y (rows).");
        addOpt_reqArg("translate-z", 803, "Translate Z (slices).");
    }

    protected boolean parseArg(int c)
    {
    	switch(c) {
			case 800:
				output_suffix = getStringArg("suffix");
				System.out.println("Output suffix is :" + output_suffix + ":");
				break;
			case 801:
				trans_x = getDoubleArg("translate X");
				break;
			case 802:
				trans_y = getDoubleArg("translate Y");
				break;
			case 803:
				trans_z = getDoubleArg("translate Z");
				break;
			default:
				return false;
		}
		return true;
    }

	final public void initParams(MCVolume v)
	{
		af  = new Affine();
	}

    final public void printArgs()
   	{
   	}

	public void preSave(MRIFile b)
	{
		comment = "affine transformation ";
		if( trans_x != 0.0 ) comment += "[translateX " + trans_x + "]";
		if( trans_y != 0.0 ) comment += "[translateY " + trans_y + "]";
		if( trans_z != 0.0 ) comment += "[translateZ " + trans_z + "]";

		super.preSave(b);
	}
		
    final public MCVolume filter(MCVolume v)
	{
		af.translateX(v, trans_x);
		af.translateY(v, trans_y);
		af.translateZ(v, trans_z);
		return v;
	}
}
