package apps.aniso;

import cj.appkit.FilterAppKit;
import cj.volume.MCVolume;
import cj.volume.measure.*;
import cj.volume.process.AnisotropicFilter;

final public class AnisoApp {
   public static void main(String argv[]) { new AnisoKit().main(argv); }
}

final class AnisoKit extends FilterAppKit {

	private float kappa = 0;
    private boolean constKappa = false;
	private boolean dim2 = true;
	private boolean channel = true;
	private int niters = 1;
	private NoiseEstimate ne = new NoiseEstimate();
	private AnisotropicFilter af = null;
    private String comment;
	private String output_suffix = "A";

	public String getAppName() { return "aniso"; }
	public String getDialogName() { return "Process image"; }
    public String getOutputSuffix() { return output_suffix; }
    public String getComment() { return comment; }

    protected void moreOpts() { 
        addOpt_reqArg("niters", 'i', "Number of iterations (def: 1).");
        addOpt_reqArg("kappa", 'k', "Noise estimate (def: calculated).");
        addOpt_noArg("2d", '2', "Do 2D filtering.");
        addOpt_noArg("3d", '3', "Do 3D filtering.");
        addOpt_noArg("per-channel", 'p', "Filter per image rather than over channels.");
        addOpt_reqArg("output-suffix", 800, "Output suffix (def: 'A').");
    }

    protected boolean parseArg(int c)
    {
    	switch(c) {
			case '2':
				dim2 = true;
				break;
			case '3':
				dim2 = false;
				break;
			case 'p':
				channel = false;
				break;
			case 800:
				output_suffix = getStringArg("suffix");
				break;
			case 'i':
				niters = getIntArg("number of iterations");
				break;
			case 'k':
				constKappa = true;
				kappa = getFloatArg("kappa");
				break;
			default:
				return false;
		}
		return true;
    }

	final public void initParams(MCVolume v)
	{
		if (!constKappa) { kappa = 2 * ne.gradientMethod( v ); }
		af = new AnisotropicFilter(kappa);
	}

    final public void printArgs()
   	{
		System.out.println("Number of iterations: " + niters);
		System.out.println("Kappa: " + kappa);
		if( channel )
		{
			System.out.println("Channel filtering: yes");
		}
		else
		{
			System.out.println("Channel filtering: no");
		}
   	}

    final public MCVolume filter(MCVolume v)
	{
		if( channel )
		{
			if( dim2 ) 
			{ 
				af.aniso2d(v, niters);
				comment = "aniso2d (channel) kappa=" + kappa + " iters=" + niters;
			}
			else 
			{
				af.aniso3d(v, niters);
				comment = "aniso3d (channel) kappa=" + kappa + " iters=" + niters;
			}
		}
		else 
		{
			if( dim2 ) 
			{ 
				af.aniso2d_perChannel(v, niters);
				comment = "aniso2d (not channel) kappa=" + kappa + " iters=" + niters;
			}
			else 
			{
				af.aniso3d_perChannel(v, niters);
				comment = "aniso3d (not channel) kappa=" + kappa + " iters=" + niters;
			}
		}

		return v;
	}
}
