package apps.autoseg;

import cj.appkit.*;
import cj.volume.*;
import cj.volume.measure.*;
import cj.volume.process.*;
import cj.roi.*;
import cj.math.convexp.*;
import cj.math.matrix.*;
import cj.fileformats.*;
import cj.utilities.*;
import java.util.Arrays.*;

final public class AutoSegApp {
    public static void main(String argv[]) { new AutoSegKit().main(argv); }
}

final class AutoSegKit extends FilterAppKit {
	boolean useGrown = false;

    public String getAppName() { return "autoseg"; }
    public String getDialogName() { return "Load dual MIF"; }
	public String getOutputSuffix() { return "_seg"; }
	public String getComment() { return "autoseg"; }

	protected void moreOpts() {
		addOpt_noArg("pre-grown",'g',"Use pre-grown lesions from ROI file");
	}
	
	protected boolean parseArg(int c) {
		if (c == 'g') return useGrown = true; else return false;
	}

    public void help() {
        System.out.println("autoseg [options] C0910D01.MIF ...");
        printOptions();
        System.out.println("Basis points are loaded from C0910S01.roi.");
    }

final public void printArgs() { }
final public void initParams(MCVolume v) { }

final public MCVolume filter(MCVolume v)
{
	MIFName mn = new MIFName(filename);
	String mifstr = mn.toString();
	String roistr = mn.recode('S').getRelatedPath(".roi");

	if (verbose > 0)
	{
		System.out.println("Reading "+roistr+"...");
	}

	ROIList rl = new ROIList(roistr);
	ConvexP cp = new ConvexP(useGrown?3:4, 2);

	String[] tl = {"wm","gm","csf","les"};

	Classify cl = new Classify(v,cp,rl,tl,useGrown);

	MCVolume out = cl.filter();

	if (verbose > 0) System.out.println("\nDone!");
	if (verbose > 1) AppKit.memStatus(System.out,true);

	return out;
}

}
