package apps.csfmask;

import java.io.*;
import cj.fileformats.*;
import cj.volume.MCVolume;
import cj.volume.measure.*;
import cj.volume.process.CSFMask;
import cj.appkit.FilterAppKit;

final public class CSFMaskApp {
   public static void main(String argv[]) {
      FilterAppKit myApp = new CSFMaskKit(); myApp.main(argv);
   }
}

final class CSFMaskKit extends FilterAppKit 
{
	private CSFMask csf = null;

    private String comment = "";
	private String output_suffix = "_csf";

	public String getAppName() { return "csfmask"; }
	public String getDialogName() { return "Process image"; }
    public String getOutputSuffix() { return output_suffix; }
    public String getComment() { return comment; }

    public String getOutputFilename(String filename)
    {
		String in = (new File(filename)).getName();
        int ext_index = in.lastIndexOf(".");

        if( ext_index > 0 && in.endsWith(".gz") ) {
            ext_index = in.lastIndexOf(".", ext_index-1);
        }

        if (ext_index < 0) return in + getOutputSuffix();
        else return in.substring(0, ext_index) + getOutputSuffix() +
                    in.substring(ext_index);
    }

    protected void moreOpts() { 
        addOpt_reqArg("output-suffix", 's', "Output suffix (def: '_aniso').");
    }

    protected boolean parseArg(int c)
    {
    	switch(c) {
			case 's':
				output_suffix = getStringArg("suffix");
				break;
			default:
				return false;
		}
		return true;
    }

	final public void initParams(MCVolume v)
	{
		csf = new CSFMask();
	}

    final public void printArgs()
   	{
   	}

    final public MCVolume filter(MCVolume v)
	{
		csf.setVerbose(verbose);
		MCVolume out = csf.csfmask(v);
		return out;
	}
}
