package apps.displaysl;

import java.awt.*;
import java.awt.event.*;
import cj.volume.*;

class MCShiftPanel extends MCVolumePanel
{
	int[] shifts;
	byte[] mySlice;

	MCShiftPanel(MCVolume v)
	{
		super(v, true);
		shifts = new int[imDepth]; java.util.Arrays.fill(shifts,0);
		mySlice = new byte[imArea];
	}

	MCShiftPanel(MCVolume v, boolean normMode)
	{
		super(v, normMode, false);
		shifts = new int[imDepth]; java.util.Arrays.fill(shifts,0);
		mySlice = new byte[imArea];
	}

	MCShiftPanel(MCVolume v, boolean normMode, boolean scaleMode)
	{
		super(v, normMode, scaleMode);
		shifts = new int[imDepth]; java.util.Arrays.fill(shifts,0);
		mySlice = new byte[imArea];
	}

	public float getData(int x, int y, int z)
	{
		int newy = (y - shifts[sliceIndex]) % imHeight;
		newy = (newy + imHeight) % imHeight;

		return super.getData(x, newy, z);
	}

	private void rowShift(int slice, int offset)
	{
		shifts[sliceIndex] += offset;

		offset = ( (offset % imHeight) + imHeight ) % imHeight;
		if (offset == 0) return;

		int bottom = offset, top = imHeight - bottom;

		System.arraycopy( scaledData, slice*imArea, mySlice, offset*imWidth, top*imWidth);
		System.arraycopy( scaledData, slice*imArea + top*imWidth, mySlice, 0, bottom*imWidth);

		System.arraycopy( mySlice, 0, scaledData, slice*imArea, imArea);
	}

	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);

		int sh = shifts[sliceIndex];

		if (sh != 0)
		{
			String s = (sh>0)?"+"+sh:""+sh;
			g.setColor(Color.green);
			g.drawString(s, getWidth()-27,10);
		}
	}

	protected boolean otherKeyTyped(KeyEvent e)
	{
		char c = (char) e.getKeyChar();

		switch(c)
		{
			case '+':
				rowShift(sliceIndex, +1);
				displaySlice(sliceIndex);
				break;

			case '-':
				rowShift(sliceIndex, -1);
				displaySlice(sliceIndex);
				break;

			case 'm':
				sliceIndex++;
				if (sliceIndex%imChannels == 0) sliceIndex-=imChannels;
				displaySlice(sliceIndex);
				break;

			default:
				return false;
		}

		return true;
	}
}
