package apps.drifter;

import java.io.*;
import cj.volume.*;

/**
 * A lightweight class for Slices with vertical shifts.
 *
 * Currently this only implements getRow(float[],int), which is the minimum
 * needed for HistBase to work.  It could be replaced by Affine once that is
 * working properly, although it is significantly faster for this purpose.
 */

public class ShiftSlice extends Slice
{
	protected int shift=0;
	protected Slice s;

	public ShiftSlice() { }
	public ShiftSlice(Slice in) { setSlice(in); }

	public void setSlice(Slice in)
	{
		s = in; shift=0;
		this.nRows = s.getNRows();
		this.nCols = s.getNCols();
	}

	public void getRow(float[] v, int row) { s.getRow(v,mapRow(row)); }

	// shift direction is down
	public void setShift(int sh) { this.shift = sh; }
	public void addShift(int sh) { this.shift += sh; }

	private int mapRow(int row)
		{ row = (row-shift)%nRows; return (row<0)?row+nRows:row; }

	public void write(DataOutput ds, final int datatype)
		{ flush(); super.write(ds, datatype); }

	/**
	 * Physically remap data array in preparation for writing.
	 */
	public void flush()
	{
		data = new float[nRows*nCols];
		float[] row = new float[nCols];

		for(int i=0;i<nRows;i++)
			{ getRow(row,i); System.arraycopy(row,0,data,i*nCols,nCols); }
	}
}
