package apps.genesis2bff;

import cj.fileformats.BFF;
import cj.fileformats.BFFfromGE;
import cj.fileformats.Genesis;
import javax.swing.*;
import java.io.*;
import java.awt.*;
import java.util.Vector;
import java.awt.image.*;
import gnu.getopt.*;

class Genesis2BFFApp {

	public static void help()
	{
		System.out.println("ge2bff [Options] <dir> [ <files> ]");
		System.out.println("Options: ");
		System.out.println("  [-h|--help] - This help.");
		System.out.println("  [-v|--verbose] - Be verbose.");
		System.out.println("  [-t|--type] <image type> - Specify image types to convert (def: all).");
		System.out.println("  [--output-suffix] <suffix> - Output suffix (def: none).");
		System.out.println("");
		System.out.println("<image type>: Can be one of real,imag,magn,phase");
		System.out.println("              Specify several types with separate -t switches.");
		System.out.println("");
	}

	public static void main(String args[]) {

		int nslices;
		String directory = "";
		String[] filenames = null;
		int verbose = 0;
		Vector type = new Vector();
		String output_suffix = "";

		//-------------------------------------------------------
		//
		//  Check the arguements to the program.
		//
		LongOpt[] longopts = new LongOpt[4];
		longopts[0] = new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h');
		longopts[1] = new LongOpt("verbose", LongOpt.NO_ARGUMENT, null, 'v');
		longopts[2] = new LongOpt("type", LongOpt.REQUIRED_ARGUMENT, null, 't');
		longopts[3] = new LongOpt("output-suffix", LongOpt.REQUIRED_ARGUMENT, null, 800);
		Getopt g = new Getopt("Genesis2BFF", args, "hvt:W", longopts);
		int c;
		while((c=g.getopt()) != -1 ) {
			switch(c) {
				case 800:
					output_suffix = g.getOptarg();
				break;
				case 't':
					type.add( g.getOptarg() );
				break;
				case 'v':
					verbose++;
				break;
				case 'h':
					help();
					System.exit(0);
				break;
			}
		}

		/*
		 *  See if there is a directory specified.
		 */
		if( g.getOptind() >= args.length ) {
			//directory = "./";
			help();
			System.exit(0);
		}
		else {
			directory = args[g.getOptind()];
		}

		/*
		 *  If there are filenames on the command lines as well,
		 *  then we need to put them into the list.
		 */
		if( g.getOptind() < args.length ) {
			filenames = new String[args.length - g.getOptind() - 1];
			for (int ii = g.getOptind()+1; ii < args.length ; ii++) {
				filenames[ii-(g.getOptind()+1)] = args[ii];
			}	

			if( verbose > 0 ) 
			{
				for (int ii = 0; ii < filenames.length ; ii++) {
					System.out.println("Filename is " + filenames[ii] );
				}
			}
		}

		//----------------------------------------------------------
		//
		//  Create the BFF
		//
		BFFfromGE b = new BFFfromGE();

		//----------------------------------------------------------
		//  
		//  Setup the data types to convert.
		//
		if( type.size() == 0 )
		{
			b.setConvertType("magn");
			b.setConvertType("phase");
			b.setConvertType("real");
			b.setConvertType("imag");
		}
		else
		{
			for(int ii=0; ii<type.size(); ii++) 
			{
				if( ((String)type.get(ii)).indexOf(",") != -1 )
				{
					System.out.print("** ge2bff: The -t switch can ");
					System.out.print("not contain a commma separated ");
					System.out.println("list of types, use a separate -t switch for each one. **");
				}

				b.setConvertType((String)type.get(ii));
			}
		}
	
		//----------------------------------------------------------
		//
		//  Do the conversion. 
		//
		Genesis ge;
		if( filenames == null || filenames.length == 0 ) {
			ge = b.fromGenesis(directory);
		}
		else {
			ge = b.fromGenesis(directory, filenames);
		}
		
		//----------------------------------------------------------
		//
		//  Write it out. 
		//
		String filename_out;
		filename_out = "ag" + ge.get_SE_se_exno();
		if( ge.get_SE_se_no() < 10 )
		{
			filename_out +=  "_00";
		}
		else if( ge.get_SE_se_no() >= 10 && ge.get_SE_se_no() < 100 )
		{
			filename_out +=  "_0";
		}
		else if( ge.get_SE_se_no() >= 100 )
		{
			filename_out +=  "_";
		}
		filename_out += ge.get_SE_se_no() + output_suffix + ".bff.gz";

		b.writeFromGenesis(filename_out);

	}
}
