package apps.histdisp;

import javax.swing.*;
import javax.swing.event.*;
import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import cj.fileformats.*;
import cj.volume.*;
import cj.volume.measure.*;
import graph.*;

class HistDispApp extends Graph2D implements KeyListener, MouseListener, MouseMotionListener {

	DataSet data1 = null;
	DataSet data2 = null;
	Axis xaxis = null,yaxis = null;
	MCVolume mcvolume = null;
	Histogram hh = null;

	public static void help() {
		System.out.println("dispmri [options] [{filename|directory}]");
		System.out.println("Options:");
		System.out.println("\t[-h | --help] - Display this help.");
		System.out.println("This will display any MIF3/MIF4/BFF/TBF or raw data file.");
	}

	public static void main(String args[]) {

		String filename = "";
		boolean help_flag = false;

		//-------------------------------------------------------
		//
		//  Check the arguements to the program.
		//
		if( args.length == 0 ) {
			filename = "./";
		}
		else {
			if( args.length == 0 ) {
				filename = args[0];
			}
			else { 

				//
				//  Parse the parameters.
				//
				int param_num=0;

				do {
					if( args[param_num].compareTo("--help") == 0 || 
					    args[param_num].compareTo("-h") == 0 ) {
						help_flag = true;
					}
					param_num++;
				} while( param_num < args.length );

				filename = args[args.length-1];
			}
		}

		if( help_flag ) {
			help();
			System.exit(0);
		}

		//-------------------------------------------------------
		//
		//  If the filename is a directory, then give the user the
		//  opportunity to find the file.
		//
        if( (new File(filename)).isDirectory() ) {
            //
            //  Bring up the dialog to get the file.
            //
            JFrame dialogFrame = new JFrame("Show image");
            dialogFrame.setFont( new Font("Helvetica", Font.PLAIN, 12 ) );

            JFileChooser fd = new JFileChooser(filename);
            fd.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
            fd.setFont( new Font("Helvetica", Font.PLAIN, 12 ) );
            int returnVal = fd.showOpenDialog(dialogFrame);                   

            //  if they hit cancel then we quit.
            if( returnVal == JFileChooser.CANCEL_OPTION ) {
                System.exit(0);
            }
			
            filename = fd.getSelectedFile().getAbsolutePath();
        }                                                     

		//-------------------------------------------------------
		//
		//  Bring up a dialog to say that we are loading.
		//
		JFrame loading_frame = new JFrame();
//		JDialog loading_dialog = new JDialog(loading_frame, "Loading " + filename);
		JTextArea loading_dialog = new JTextArea("Loading " + filename);
		loading_dialog.setFont(new Font("helvetica", Font.BOLD, 14));
		loading_frame.getContentPane().add( loading_dialog ); 
		loading_frame.setResizable(false);
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		loading_frame.setSize(300,50);
		loading_frame.setLocation(screenSize.width/2-150,screenSize.height/2-25);
		loading_frame.pack();
		loading_frame.show();

		//-------------------------------------------------------
		//
		//  Load up the file.
		//
		MRIFile b = MRIFile.readMRIFile(filename);

		//-------------------------------------------------------
		//
		//  Destroy the dialog to say that we are loading.
		//
		loading_frame.dispose();
		loading_frame = new JFrame(filename);
		
		loading_frame.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );

		HistDispApp dm = new HistDispApp(b.getMCVolume(), 113);
		loading_frame.getContentPane().add(dm);
		loading_frame.addKeyListener(dm);
		loading_frame.addMouseListener(dm);
		loading_frame.addMouseMotionListener(dm);
		loading_frame.setBackground(Color.white);
		loading_frame.setResizable(true);
		loading_frame.setSize(400,200);
		loading_frame.show();
	}

	public HistDispApp(MCVolume mm_in, int nbins)
	{
		hh = new Histogram();

		mcvolume = mm_in;	

		initHistogram(nbins);
	}

	public HistDispApp(MCVolume mm_in)
	{
		hh = new Histogram();

		mcvolume = mm_in;	

		initHistogram(-1);
	}

	public void initHistogram(int nbins)
	{
		System.out.println("Called initHistogram with " + nbins + " bins");
		if( nbins == -1 ) {
			hh.histogram(mcvolume, 200, Integer.MAX_VALUE);
		}
		else {
			hh.histogram(mcvolume, 200, Integer.MAX_VALUE,  nbins);
		}

		float[] bins = hh.getBins();
		float[] counts = hh.getCounts();
		double[] data = new double[ 2*bins.length ];

		int ii, jj;
		for(ii=0,jj=0; ii<bins.length; jj+=2,ii++) {
			data[jj] = (double)bins[ii];
			data[jj+1] = (double)counts[ii];
		}

		setPreferredSize(new Dimension( 400, 200 ) );
		drawzero = true;
		drawgrid = true;
		borderTop = 20;
		borderRight = 40;
		setBackground( Color.white );

		data1 = loadDataSet(data,bins.length);
		data1.linecolor = new Color(0,0,0);

		data2 = loadDataSet(data,bins.length);
		data2.linestyle = 0;
		data2.marker    = 1;
		data2.markerscale = 1.5;
		data2.markercolor = new Color(0,0,0);

		xaxis = createAxis(Axis.BOTTOM);
		xaxis.attachDataSet(data1);
		xaxis.attachDataSet(data2);
		xaxis.setTitleText("Bins");
		xaxis.setTitleFont(new Font("TimesRoman",Font.PLAIN,12));
		xaxis.setLabelFont(new Font("Helvetica",Font.PLAIN,10));

		yaxis= createAxis(Axis.LEFT);
		yaxis.attachDataSet(data1);
		yaxis.attachDataSet(data2);
		yaxis.setTitleText("Counts");
		yaxis.setTitleFont(new Font("TimesRoman",Font.PLAIN,12));
		yaxis.setLabelFont(new Font("Helvetica",Font.PLAIN,10));

	}


	public void mouseDragged(MouseEvent e) { } 
	public void mouseMoved(MouseEvent e) { } 
    public void mouseReleased(MouseEvent e) {}
    public void mousePressed(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseClicked(MouseEvent e) { } 

	public void keyReleased(KeyEvent e) {}
	public void keyPressed(KeyEvent e) {}

	/**
	 *  Overloaded function of what to do when a key is typed.
	 *
	 *  @param e - keyEvent 
	 *  @return none
	 */
	public void keyTyped(KeyEvent e)
	{
		boolean unknownKey = false;
		char c = (char) e.getKeyChar();
			
		switch(c) {
			case '+':
				initHistogram(hh.getBins().length + 10);
			break;
			case '-':
				initHistogram(hh.getBins().length - 10);
			break;
			case 'q':
				System.exit(0);
			break;
		}
	}
}
