package apps.idmask;

import java.io.File;
import cj.fileformats.*;
import cj.volume.MCVolume;
import cj.volume.measure.*;
import cj.volume.process.IDMask;
import cj.appkit.FilterAppKit;

final public class IDMaskApp {
   public static void main(String argv[]) {
      IDMaskKit myApp = new IDMaskKit(); myApp.main(argv);
   }
}

final class IDMaskKit extends FilterAppKit 
{
	private IDMask id = null;

    private String comment = "";
	private String output_suffix = "_id";

	public String getAppName() { return "idmask"; }
	public String getDialogName() { return "Process image"; }
    public String getOutputSuffix() { return output_suffix; }
    public String getComment() { return comment; }

    public String getOutputFilename(String filename)
    {
		String in = (new File(filename)).getName();
        int ext_index = in.lastIndexOf(".");

        if( ext_index > 0 && in.endsWith(".gz") ) {
            ext_index = in.lastIndexOf(".", ext_index-1);
        }

        if (ext_index < 0) return in + getOutputSuffix();
        else return in.substring(0, ext_index) + getOutputSuffix() +
                    in.substring(ext_index);
    }

	private double gaussian_factor = 0.4;

    public void help() {
		System.out.println("idmask [options] <filename> [<filename> ...]");
		printOptions();
    }

    protected void moreOpts() { 
        addOpt_reqArg("output-suffix", 's', "Output suffix (def: '_aniso').");
        addOpt_reqArg("gaussian-factor", 'g', "Gaussian factor (def: 0.4).");
    }

    protected boolean parseArg(int c)
    {
    	switch(c) {
			case 'g':
				gaussian_factor = getDoubleArg("Gaussian factor");
				break;
			case 's':
				output_suffix = getStringArg("suffix");
				break;
			default:
				return false;
		}
		return true;
    }

	final public void initParams(MCVolume v)
	{
		id = new IDMask();
		id.setVerbose( verbose );
		id.setGaussianFactor( gaussian_factor );
	}

    final public void printArgs()
   	{
   	}

    final public MCVolume filter(MCVolume v)
	{
		MCVolume out = id.idmask(v);
		return out;
	}
}
