package apps.median;

import cj.volume.*;
import cj.volume.measure.*;
import cj.volume.process.*;
import cj.appkit.FilterAppKit;

final public class MedianApp {
   public static void main(String argv[]) { new MedianKit().main(argv); }
}

final class MedianKit extends FilterAppKit {

	private int niters = 1;
	private boolean process2D = true;
	private String output_suffix = "M";

	private MedianFilter mf = null;
    private String comment = "";

	public String getAppName() { return "median"; }
	public String getDialogName() { return "Process image"; }
    public String getOutputSuffix() { return output_suffix; }
    public String getComment() { return comment; }

    protected void moreOpts() { 
        addOpt_noArg("2d", '2', "Process slice by slice (def: true).");
        addOpt_noArg("3d", '3', "Process cubes (def: false).");
        addOpt_reqArg("niters", 'i', "Number of iterations (def: 1).");
        addOpt_reqArg("output-suffix", 800, "Output suffix (def: 'M').");
    }

    protected boolean parseArg(int c)
    {
    	switch(c) {
			case '2':
				process2D = true;
				break;
			case '3':
				process2D = false;
				break;
			case 'i':
				niters = getIntArg("number of iterations");
				break;
			case 800:
				output_suffix = getStringArg("suffix");
				break;
			default:
				return false;
		}
		return true;

    }

	final public void initParams(MCVolume v)
	{
		mf = new MedianFilter();
	}

    final public void printArgs()
   	{
		System.out.println("Number of iterations: " + niters);
   	}

    final public MCVolume filter(MCVolume v)
	{
		if( process2D )
			mf.process2D(v, niters);
		else
			throw new RuntimeException("3rd not supported."); //mf.process3D(v, niters);

		comment = "median";
		comment += " iters=" + niters;

		return v;
	}
}
