package apps.register;

import java.util.StringTokenizer;
import java.io.*;
import cj.fileformats.*;
import cj.utilities.*;
import cj.volume.*;
import cj.volume.process.register.*;
import cj.math.interpolate.*;
import cj.appkit.*;

final public class RegisterApp {
   public static void main(String argv[]) { new RegisterKit().main(argv); }
}

final class RegisterKit extends FilterAppKit 
{
	protected boolean first_scan = true;

	private String basename = null;
	private String output_suffix = "R";
	private String comment = "";
	private Register registerer = null;
	private String similarity = "woods";
	private double ftol = 1e-6;

	protected double[] vx1 = new double[3], vx2 = new double[3];

	private Volume stat;

	//private int rowSkip = 5;
	//private int colSkip = 5;
	//private int sliceSkip = 1;
	private int[] rowSkip = {5};
	private int[] colSkip = {5};
	private int[] sliceSkip = {1};

	// Output dimensions
	private int odx = -1;
	private int ody = -1;
	private int odz = -1;

	private String rdf_file = "";

	private String xInternalInterp = "linear";
	private String yInternalInterp = "linear";
	private String zInternalInterp = "linear";

	private String xInterp = "sinc";
	private String yInterp = "sinc";
	private String zInterp = "linear";

	private double cjinit[] = new double[6];

	public String getAppName() { return "register"; }
	public String getDialogName() { return "Process image"; }
    public String getOutputSuffix() { return output_suffix; }
    public String getComment() { return comment; }

	public void help()
	{
        System.out.println(getAppName() + " [options] <static filename> <dynamic filename> [<dynamic filename2> ...]");
		printOptions();

		System.out.println("Interplator can be one of 'linear', 'cubic', or 'sinc'.");
	}

    protected void moreOpts() { 
        addOpt_reqArg("similarity", 801, "Similarity measure ('woods', 'l1', 'shannon', 'l2', def: 'woods').");

        addOpt_reqArg("row-skip", 804, "Rows to skp (def: 5).");
        addOpt_reqArg("col-skip", 805, "Columns to skip (def: 5).");
        addOpt_reqArg("slice-skip", 806, "Slices to skip (def: 1).");

        addOpt_reqArg("ix", 807, "Interpolator in X direction (def: sinc).");
        addOpt_reqArg("iy", 808, "Interpolator in Y direction (def: sinc).");
        addOpt_reqArg("iz", 809, "Interpolator in Z direction (def: linear).");

        addOpt_reqArg("internal-interpolator-x", 810, "Internal X interpolator (def: linear).");
        addOpt_reqArg("internal-interpolator-y", 811, "Internal Y interpolator (def: linear).");
        addOpt_reqArg("internal-interpolator-z", 812, "Internal Z interpolator (def: linear).");

        addOpt_reqArg("cjinit", 813, "Initial parameters rx ry rz tx ty tz (def: 0 0 0 0 0 0 ).");
        addOpt_reqArg("ftol", 't', "Function tolerance (def: 1e-6).");
    }

    protected boolean parseArg(int c)
    {
		String temp;
		StringTokenizer st = null;

    	switch(c) {
			case 't':
				ftol = getDoubleArg("Function Tolerance");
			break;

			case 801:
				similarity = getStringArg("Similarity measurer");
			break;

			case 804:
				temp = getStringArg("row skip string");

				st = new StringTokenizer(temp, ":");

				rowSkip = new int[ st.countTokens() ];
				for(int ii=0; ii<rowSkip.length; ii++) 
				{
					rowSkip[ii] = Integer.parseInt(	st.nextToken() );

					if( rowSkip[ii] == 0 ) 
					{
						System.out.println("Huh? The row can't have zero skips!!");
						System.exit(-1);
					}
				}

			break;
			case 805:
				temp = getStringArg("column skip string");

				st = new StringTokenizer(temp, ":");

				colSkip = new int[ st.countTokens() ];
				for(int ii=0; ii<colSkip.length; ii++) 
				{
					colSkip[ii] = Integer.parseInt(	st.nextToken() );

					if( colSkip[ii] == 0 ) 
					{
						System.out.println("Huh? The column can't have zero skips!!");
						System.exit(-1);
					}
				}
			break;
			case 806:
				temp = getStringArg("slice skip string");

				st = new StringTokenizer(temp, ":");

				sliceSkip = new int[ st.countTokens() ];
				for(int ii=0; ii<sliceSkip.length; ii++) 
				{
					sliceSkip[ii] = Integer.parseInt(	st.nextToken() );

					if( sliceSkip[ii] == 0 ) 
					{
						System.out.println("Huh? The slices can't have zero skips!!");
						System.exit(-1);
					}
				}
			break;

			case 807:
				xInterp = getStringArg("interpolator X");
			break;
			case 808:
				yInterp = getStringArg("interpolator Y");
			break;
			case 809:
				zInterp = getStringArg("interpolator Z");
			break;

			case 810:
				xInternalInterp = getStringArg("Internal Interpolator X");
			break;
			case 811:
				xInternalInterp = getStringArg("Internal Interpolator Y");
			break;
			case 812:
				xInternalInterp = getStringArg("Internal Interpolator Z");
			break;

			case 813:
				for(int ii=0; ii<6; ii++) 
					cjinit[ii] = getDoubleArg("Yo Parameter " + ii);
			break;

			case 'r':
				rdf_file = getStringArg("RDF File");
			break;
			default:
				return false;
		}
		return true;

    }

	/**
	 *  Fill this one in!!
	 */
	final public void pre()
	{
		filterMode = READ_ONLY;

		registerer = new RegisterRigid();
		registerer.setVerbose(verbose);

		((RegisterRigid)registerer).setTolerance(ftol);

		/*
		 *  Check to make sure that there are the same number
		 *  of skips in each list.
		 */

		if( rowSkip.length != colSkip.length || 
			rowSkip.length != sliceSkip.length )
		{
			System.out.println("register:  Number of rows, columns and slices to skip must be the same.");
			System.exit(-1);
		}

		((RegisterRigid)registerer).setSimilarityMeasure( similarity );
		((RegisterRigid)registerer).setRowSkip( rowSkip );
		((RegisterRigid)registerer).setColSkip( colSkip );
		((RegisterRigid)registerer).setSliceSkip( sliceSkip );
	}

	final public void initParams(MRIFile b)
    {
		if (first_scan)
		{
			vx1[0] = (double)((Analyze)b).GetPixelX()/1000.0;
			vx1[1] = (double)((Analyze)b).GetPixelY()/1000.0;
			vx1[2] = (double)((Analyze)b).GetSliceThick()/1000.0;
		}
		else
		{
			vx2[0] = (double)((Analyze)b).GetPixelX()/1000.0;
			vx2[1] = (double)((Analyze)b).GetPixelY()/1000.0;
			vx2[2] = (double)((Analyze)b).GetSliceThick()/1000.0;
		}
    }

	final public void preSave(MRIFile b)
	{
        ((Analyze)b).SetPixelX((int)(1000.0*vx1[0]));
        ((Analyze)b).SetPixelY((int)(1000.0*vx1[1]));
        ((Analyze)b).SetSliceThick((int)(1000.0*vx1[2]));
	}

    public void printArgs()
	{
		if (!first_scan)
			{ System.out.println("Registering " + filename + " to " + basename); }
	}

    final public MCVolume filter(MCVolume v)
	{
		if (first_scan)
		{
			first_scan = false;
			basename = new String(filename);
			stat = v.getVolume(0);
			return v;
		}

		// This has to occur AFTER the first scan has been skipped
		filterMode = READ_WRITE;

		registerer.setInternalInterpolatorX( InterpolateRow.createInterpolator( xInternalInterp ) );
		registerer.setInternalInterpolatorY( InterpolateRow.createInterpolator( yInternalInterp ) );
		registerer.setInternalInterpolatorZ( InterpolateRow.createInterpolator( zInternalInterp ) );

		registerer.setInterpolatorX( InterpolateRow.createInterpolator( xInterp ) );
		registerer.setInterpolatorY( InterpolateRow.createInterpolator( yInterp ) );
		registerer.setInterpolatorZ( InterpolateRow.createInterpolator( zInterp ) );
		
		((RegisterRigid)registerer).setInitialRotationX(cjinit[0]);
		((RegisterRigid)registerer).setInitialRotationY(cjinit[1]);
		((RegisterRigid)registerer).setInitialRotationZ(cjinit[2]);
		((RegisterRigid)registerer).setInitialTranslationX(cjinit[3]);
		((RegisterRigid)registerer).setInitialTranslationY(cjinit[4]);
		((RegisterRigid)registerer).setInitialTranslationZ(cjinit[5]);

		Volume out = registerer.register(stat, vx1, v.getVolume(0), vx2);

        MCVolume mv = new MCVolume(out.getNRows(),out.getNCols(),out.getNSlices(),1);
		mv.setVolume(out,0);

		return mv;
	}
}
