import javax.swing.*;
import javax.swing.event.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

class DisplayMIF extends JFrame implements KeyListener, MouseListener, MouseMotionListener {

	private ImagePanel imagePanel;
	private JSlider sliceSlider;
	private int width, height, nslices;
	private int[] imageData;
	private Point mouseLoc;

	//-----------------------------------------------------
	//
	//  Constructor
	//
	DisplayMIF(String title, int[] id, int w, int h, int slices) {

		width = w;
		height = h;
		nslices = slices;

		//
		//  Copy the data over (scaled).
		//
		int max=0;
		for(int ii=0; ii<width*height*nslices; ii++) {
			if( id[ii] > max ) {	
				max = id[ii];
			}
		}
		System.out.println("Maximum value is " + max);

		//
		//  Scale and set the imageData variable.
		//
		imageData = new int[width*height*nslices];
		for(int ii=0; ii<width*height*nslices; ii++) {
			imageData[ii] = (int)((float)id[ii] / (float)max * 255.0);
		}

		//
		//  Set the true imageData variable to RGB.
		//
		int jj;
		for(int ii=0; ii<width*height*nslices; ii++) {
			jj = imageData[ii] & 0xff;
			imageData[ii] = ((255<<24) | (jj<<16) | (jj<<8) | jj);
		}

		
		//
		//  Set the mouse and key listeners.
		//
		addKeyListener(this);
		addMouseListener(this);
		addMouseMotionListener(this);

		getContentPane().setLayout(new BorderLayout(2,1));

		//
		//  Create the panel.
		//
		imagePanel = new ImagePanel(imageData, width, height, nslices);
		imagePanel.setSize(width, height);
		imagePanel.setPreferredSize(new Dimension(width, height));
		getContentPane().add(imagePanel, BorderLayout.NORTH);
		imagePanel.setBorder(BorderFactory.createTitledBorder("Image"));


		
		//
		//  Create the slider.
		//
		sliceSlider = new JSlider(JSlider.HORIZONTAL, 1, nslices, 1);
		imagePanel.setSize(256,10);
		getContentPane().add(sliceSlider, BorderLayout.SOUTH);
		sliceSlider.setBorder(BorderFactory.createTitledBorder("Slice Number"));

		sliceSlider.setMajorTickSpacing(5);
		sliceSlider.setMinorTickSpacing(1);
		sliceSlider.setPaintTicks(true);
		sliceSlider.setPaintLabels(true);
		sliceSlider.addChangeListener(new SliceListener());

		//
		//  Set the size and display the frame.
		//
		setSize(260,330);
		setTitle(title);
		show();
	}

	protected void paintComponent(Graphics g) {
		imagePanel.paintComponent(g);
	}

	//-----------------------------------------------------
	//
	//  Mouse Listeners
	//
    public void mouseReleased(MouseEvent e) {}
    public void mousePressed(MouseEvent e) {
        mouseLoc = e.getPoint();
        System.out.println("Mouse Pressed (" + mouseLoc.x + ", " +mouseLoc.y+").");
    }

    public void mouseExited(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseClicked(MouseEvent e) {}

	public void mouseDragged(MouseEvent e) {}

	public void mouseMoved(MouseEvent e) {
		mouseLoc = e.getPoint();
	}

	public void keyPressed(KeyEvent e) {}
	public void keyReleased(KeyEvent e) {}
	public void keyTyped(KeyEvent e) {
        System.out.println("The " + e.getKeyChar() + " key pressed (" + mouseLoc.x + ", " +mouseLoc.y+").");

		if( e.getKeyChar() == 'q' || e.getKeyChar() == 'Q' ) {
			System.exit(0);
		}
	}

class SliceListener implements ChangeListener {
	
	public void stateChanged(ChangeEvent e) {
		JSlider source = (JSlider)e.getSource();

		if( !source.getValueIsAdjusting() ) {
			System.out.println("Slider changed to slice " + source.getValue());
			imagePanel.displaySlice(source.getValue());
		}
	}
}
}

class ImagePanel extends JPanel {

	private MemoryImageSource mi;
	private Image m_image;
	int width, height, nslices;
	int imageData[];

	ImagePanel(int[] p_imageData, int p_width, int p_height, int p_nslices) {

		//
		//  Copy the width and height to internal variables.
		//
		width = p_width;
		height = p_height;
		nslices = p_nslices;

		imageData = new int[width*height*nslices];

		//
		//  Copy the data over.
		//
		for(int ii=0; ii<width*height*nslices;ii++) {
			imageData[ii] = p_imageData[ii];
		}

		mi = new MemoryImageSource(width, height, imageData, 0, width);
		displaySlice(1);
		mi.setAnimated(true);
	}

	public void displaySlice(int sliceNumber) {

		System.out.println("Want to display slice " + sliceNumber);
	
		mi.newPixels(imageData, getColorModel(), 
			(sliceNumber-1)*width*height, width);
		System.out.println("Going to create new image");
		m_image = createImage(mi);
	}

	public void paintComponent(Graphics g) {
		g.drawImage(m_image, 0, 0, width, height, this);
	}
}
