      SUBROUTINE NNLS(A, MDIM, M, N, B, X, RNORM, W, ZZ, INDEX, MODE) 
c      IMPLICIT DOUBLE PRECISION (A-H,O-Z) 
C$$$$$$ CALLS H12,G1,G2 
C NON-NEGATIVE LEAST-SQUARES FITTING ROUTINE.  GIVEN BY LAWSON + HANSON 
C  IN  SOLVING LEAST SQUARES PROBLEMS ,PRENTICE-HALL 1974. 
C  GIVEN  MATRIX  A  (M BY N), AND AN M-VECTOR  B  , FINDS THE N-VECTOR 
C  X  THAT SOLVES THE PROBLEM 2-NORM(A*X - B) MINIMUM WITH X.GE.0 
C 
C  A(),MDIM,M,N  A  IS THE ARRAY, WITH MROWS,  N  COLS.  MDIM  IS THE 
C     ACTUAL DIMENSION GIVEN  A  IN THE CALLING PROGRAM (MDIM.GE.M) 
C     A  IS DESTROYED BY PROGRAM. 
C  B()  ON ENTRY MUST CONTAIN M-VECTOR  B.  DESTROYED BY PROGRAM 
C  X()  SOLUTION VECTOR  X.  NEED NOT BE INITIALIZED. 
C  RNORM  MINIMUM TWO-NORM. 
C  W()  AN ARRAY OF WORKING SPACE.  MUST BE DIMENSIONED AT LEAST N 
C  ZZ()  ANOTHER WORKING ARRAY.  MUST BE DIMENSIONED AT LEAST M. 
C  INDEX  ANOTHER WORKING ARRAY. MUST BE DIMENSIONED AT LEAST  N. 
C  MODE  A FLAG INDICATENG OUTCOME OF SUBROUTINE. 
C     MODE=1   MEANS ROUTINE WORKED PROPERLY. 
C     MODE=2   MEANS DIMENSIONS OF ARRAYS WERE BAD (M.LE0 .OR. N.LE.0) 
C     MODE=3   MEANS ITERATION COUNT EXCEEDED (MORE THAN 3*N ITERATI
C 
	  INTEGER MDIM, M, N
      DOUBLE PRECISION A(MDIM,N),B(M),X(N),W(N),ZZ(M),DUMMY(1) 
	  INTEGER INDEX(N)
	  DOUBLE PRECISION FACTOR, ZERO, ONE, TWO, SM, WMAX, RNORM, UP
	  DOUBLE PRECISION CC, SS, ALPHA, T, ASAVE, ZTEST, TEMP, UNORM
	  INTEGER  MODE, ITER, ITMAX, I, IZ2, IZ1, NSETP, NPP1
	  INTEGER IZMAX, IZ, L, J, II, JJ, IP, NEXT, JZ
	  EXTERNAL G1, G2, H12
C 
      ZERO=0. 
      ONE=1. 
      TWO=2. 
      FACTOR=0.01 
C 
      MODE=1 
      IF (M.GT.0 .AND. N.GT.0) GO TO 10 
      MODE=2 
      RETURN 
 10   ITER=0 
      ITMAX=3*N 
C 
C 
C  INITIALIZE ARRAYS INDEX  AND  X 
      DO 20 I=1,N 
      X(I)=ZERO 
 20   INDEX(I)=I 
C 
      IZ2=N 
      IZ1=1 
      NSETP=0 
      NPP1=1 
C     C  MAIN LOOP BEGINS HERE 
 30   CONTINUE 
C 
C  QUIT IF ALL COEFFS ARE ALREADY IN SOLUTION. OR IF M COLS OF A HAVE 
C  BEEN TRIANGULARIZED. 
      IF (IZ1.GT.IZ2 .OR. NSETP.GE.M) GO TO 350 
C 
C 
C  COMPUT COMPONENTS OF DUAL VECTOR W 
      DO 50 IZ=IZ1,IZ2 
      J=INDEX(IZ) 
      SM=ZERO 
      DO 40 L=NPP1,M 
 40   SM=A(L,J)*B(L) + SM 
 50   W(J)=SM 
C  FIND LARGEST +VE W(J) 
 60   WMAX=ZERO 
      DO 70 IZ=IZ1,IZ2 
      J=INDEX(IZ) 
      IF (W(J).LE.WMAX) GO TO 70 
      WMAX=W(J) 
      IZMAX=IZ 
 70   CONTINUE 
C 
C 
C 
C  IF WMAX.LE.0 GO TO TERMINATION (KUHN-TUCKER CONDITIONS OK) 
      IF (WMAX) 350,350,80 
 80   IZ=IZMAX 
      J=INDEX(IZ) 
C 
C 
C 
C  DSIGN OF W(J) IS OK FOR J TO BE MOVED TO SET  P. 
C  BEGIN HOUSE-HOLDER TRANS. CHECK NEW DIAGONAL ELEMENT TO AVOID NEAR 
      ASAVE=A(NPP1,J) 
      CALL H12(1, NPP1, NPP1+1, M, A(1,J), 1, UP, DUMMY, 1, 1, 0) 
      UNORM=ZERO 
      IF (NSETP.EQ.0) GO TO 100 
      DO 90 L=1,NSETP 
 90   UNORM=A(L,J)**2 + UNORM 
 100  UNORM=DSQRT(UNORM) 
      TEMP=UNORM+DABS(A(NPP1,J))*FACTOR 
      IF (TEMP-UNORM) 130,130,110 
C  COL J IS SUFFICIENTLY INDEP. COPY  B  TO ZZ. UPDATE  Z  AND SOLVE 
C  FOR  ZTEST (=PROPOSED NEW VAL FOR X(J)) 
C 
 110  DO 120 L=1,M 
 120  ZZ(L)=B(L) 
      CALL H12(2, NPP1, NPP1+1, M, A(1,J), 1, UP, ZZ, 1, 1, 1) 
      ZTEST=ZZ(NPP1)/A(NPP1,J) 
C 
C 
C              SEE IF  ZTEST IS +VE 
      IF (ZTEST) 130,130,140 
C 
C  REJECT  J  AS A CANDIDATE TO BE MOVED FROM SET  Z  TO SET  P . 
C  RESTORE A(NPP1,J), SET W(J)=0. AND LOOP BACK TO TEST DUAL 
C  COEFFFS AGAIN. 
C 
 130  A(NPP1,J)=ASAVE 
      W(J)=ZERO 
      GO TO 60 
C 
C  THE INDEX J=INDEX(IZ) HAS BEEN SELECTED TO BE MOVED FTOM SET  Z 
C  TO SET  P .  UPDATE B,  UPDATE INDICES, APPLY HOUSEHOLDER TRANS 
C  TO COLS IN NEW SET Z.  ZERO SUBDIAGONAL  ELTS IN COL  J, SET 
C  W(J)=0. 
C 
 140  DO 150 L=1,M 
 150  B(L)=ZZ(L) 
C 
      INDEX(IZ)=INDEX(IZ1) 
      INDEX(IZ1)=J 
      IZ1=1 + IZ1 
      NSETP=NPP1 
      NPP1=1 + NPP1 
C 
      IF (IZ1.GT.IZ2) GO TO 170 
      DO 160 JZ=IZ1,IZ2 
      JJ=INDEX(JZ) 
 160  CALL H12(2, NSETP, NPP1, M, A(1,J), 1, UP, A(1, JJ), 1, MDIM, 1) 
 170  CONTINUE 
C 
      IF  (NSETP.EQ.M) GO TO 190 
      DO 180 L=NPP1,M 
 180  A(L,J)=ZERO 
 190  CONTINUE 
C 
      W(J)=ZERO 
C  SOLVE TRIANGULAT SYSTEM. STORE SOL IN ZZ,TEMPORARILY 
C 
      NEXT=1 
      GO TO 400 
 200  CONTINUE 
C 
C 
C 
C 
C  SECONDARY LOOP BEFINS HERE 
 210  ITER=1 + ITER 
      IF (ITER.LE.ITMAX) GO TO 220 
      MODE=3 
C  WRITE STATEMENT HERE DELETED********************* 
      GO TO 350 
 220  CONTINUE 
C 
C 
C 
C  SEE IF ALL NEW CONSTRAINED COEFFS ARE FEASIBLE. IF NOT, FIND ALPHA 
      ALPHA=TWO 
      DO 240 IP=1,NSETP 
      L=INDEX(IP) 
      IF (ZZ(IP)) 230,230,240 
C 
 230  T=-X(L)/(ZZ(IP) - X(L)) 
      IF (ALPHA.LE.T) GO TO 240 
      ALPHA=T 
      JJ=IP 
 240  CONTINUE 
C  IF ALL NEW CONSTRAINED COEFFS ARE FEASIBLE, ALPHA IS STILL 2. IF S
C  EXIT INTO MAIN LOOP. 
C 
C 
      IF (ALPHA.EQ.TWO) GO TO 330 
C  OTHERWISE ALPHA WILL BE IN (0,1) TO INTERPOLATE BETWEEN OLDX 
C 
C 
C 
      DO 250 IP=1,NSETP 
      L=INDEX(IP) 
 250  X(L)=ALPHA*(ZZ(IP)-X(L)) + X(L) 
C 
C  MODIFY  A  BB  AND THE INDEX ARRAYS TO MOVE COEFF  I 
C  FROM  SET  P  TO SET  Z. 
C 
      I=INDEX(JJ) 
 260  X(I)=ZERO 
C 
      IF (JJ.EQ.NSETP) GO TO 290 
      JJ=1 + JJ 
      DO 280 J=JJ,NSETP 
      II=INDEX(J) 
      INDEX(J-1)=II 
      CALL G1(A(J-1,II), A(J,II), CC, SS, A(J-1,II)) 
      A(J,II)=ZERO 
      DO 270 L=1,N 
      IF (L.NE.II) CALL G2(CC, SS, A(J-1,L), A(J,L)) 
 270  CONTINUE 
 280  CALL G2(CC, SS, B(J-1), B(J)) 
 290  NPP1=NSETP 
      NSETP=NSETP - 1 
      IZ1=IZ1 -     1 
      INDEX(IZ1)=I 
C  ALL COEFFS IN SET P SHOULD BE FEASIBLE.  IF THEY ARE NOT
C  TO ROUND-OFF.  NON-POSITIVE ONES SET TO 0 AND MOVED TO . 
C 
C 
C 
C 
C 
      DO 300 JJ=1,NSETP 
      I=INDEX(JJ) 
      IF (X(I)) 260,260,300 
 300  CONTINUE 
C 
C 
C  COPY  B  TO  ZZ  AND SOLVE AGAIN AND LOOP BACK. 
      DO 310 I=1,M 
 310  ZZ(I)=B(I) 
      NEXT=2 
      GO TO 400 
 320  CONTINUE 
      GO TO 210 
C 
C  END OF SECONDARY LOOP 
 330  DO 340 IP=1,NSETP 
      I=INDEX(IP) 
 340  X(I)=ZZ(IP) 
C  ALL NEW COEFFS ARE +VE.  LOOP BACK TO BEGINNING. 
      GO TO 30 
C  END OF MAIN LOOP 
C  TERMINATING SECTION 
C 
C 
C 
C 
 350  SM=ZERO 
      IF (NPP1.GT.M) GO TO 370 
      DO 360 I=NPP1,M 
 360  SM=B(I)**2 + SM 
      GO TO 390 
 370  DO 380 J=1,N 
 380  W(J)=ZERO 
 390  RNORM=DSQRT(SM) 
      RETURN 
C 
C  SOLUTION OF LINEAR TRIANGULAR SYSTEM 
C 
C 
 400  DO 430 L=1,NSETP 
      IP=NSETP + 1 - L 
      IF (L.EQ.1) GO TO 420 
      DO 410 II=1,IP 
 410  ZZ(II)=ZZ(II) - A(II,JJ)*ZZ(IP+1) 
 420  JJ=INDEX(IP) 
 430  ZZ(IP)=ZZ(IP)/A(IP,JJ) 
      GO TO (200,320),NEXT 
C ******** FORMAT DELETED FOR PRINT STATEMENT HERE **************** 
      END
      SUBROUTINE G2(COSA,SINA,X,Y) 
C      IMPLICIT DOUBLE PRECISION (A-H,O-Z) 
	  DOUBLE PRECISION COSA, SINA, X, Y, Z
C$$$$ CALLS NO OTHER ROUTINES 
C  ROTATES A 2-VECTOR (X,Y) BY ANGLE A. OVER-WRITES ORIGINAL VECTOR 
C 
      Z=COSA*X + SINA*Y 
      Y=COSA*Y - SINA*X 
      X=Z 
      RETURN 
      END
      SUBROUTINE G1(A, B, COSA, SINA, SIG) 
C      IMPLICIT DOUBLE PRECISION (A-H,O-Z) 
C$$$$$ CALLS NO OTHER ROUTINES 
C 
C 
C 
C 
C  FROM LAWSON+HANSON -SOLVING LEAST SQUARES PROBLEMS- 1974 P 309. 
C  COMPUTES AN ORTHOGONAL MATRIX TO ROTATE (A,B) INTO (SQRT(A*A+B*B
C  MATRIX IN FORM (COSA, SINA/-SINA, COSA) AND SIG HOLDS MAGNITUDE OF 
C  (A,B).  SIG IS ALLOWED TO OVERWRITE A OR B IN CALLING ROUTINE. 
C 
      DOUBLE PRECISION A, B, COSA, SINA, SIG
      DOUBLE PRECISION ZERO, ONE, XR, YR
      ZERO=0.0 
      ONE=1.0 
      IF (DABS(A) .LE. DABS(B)) GO TO 10 
      XR=B/A 
      YR=DSQRT(ONE + XR**2) 
      COSA=DSIGN(ONE/YR, A) 
      SINA=COSA*XR 
      SIG=DABS(A)*YR 
      RETURN 
 10   IF (B) 20,30,20 
 20   XR=A/B 
      YR=DSQRT(ONE + XR**2) 
      SINA=DSIGN(ONE/YR, B) 
      COSA=SINA*XR 
      SIG=DABS(B)*YR 
      RETURN 
 30   SIG=ZERO 
      COSA=ZERO 
      SINA=ONE 
      RETURN 
      END 
      SUBROUTINE H12(MODE, LPIVOT, L1, M, U, IUE, UP, C, ICE, ICV, NCV) 
C      IMPLICIT DOUBLE PRECISION (A-H,O-Z) 
      DOUBLE PRECISION  U, UP, C, CL, SM, SM1, CLINV, B
	  INTEGER MODE, LPIVOT, L1, M, IUE, ICE, ICV, NCV, MDIM
	  INTEGER I1, I2, I3, I4, I, J, INCR
C$$$$$ CALLS NO OTHER ROUTINES 
C  CONSTRUCTION AND APPLICATION OF HOUSEHOLDER TRANSFORMATION FOR
C  MODIFIED FROM  H12  IN LAWSON+HANSEN - SOLVING LEAST SQUARES 
C  - PRENTICE-HALL 1974 (PP308,309).  DOUBLE PRECISION THROUGHOUT 
      DOUBLE PRECISION U(IUE,M),C(1) 
      IF (0.GE.LPIVOT .OR. LPIVOT.GE.L1 .OR. L1.GT.M) RETURN 
      CL=DABS(U(1,LPIVOT)) 
      IF (MODE.EQ.2) GO TO 60 
C  CONSTRUCT TRANSFORMATION 
      DO 10 J=L1,M 
 10   CL=DMAX1(DABS(U(1,J)),CL) 
      IF (CL) 130,130,20 
 20   CLINV=1.0/CL 
      SM=(U(1,LPIVOT)*CLINV)**2 
      DO 30 J=L1,M 
 30   SM=(U(1,J)*CLINV)**2 + SM 
      SM1=SM 
      CL=DSQRT(SM1)*CL 
      IF (U(1,LPIVOT)) 50,50,40 
 40   CL=-CL 
 50   UP=U(1,LPIVOT) - CL 
      U(1,LPIVOT)=CL 
      GO TO 70 
C  APLLY TRANSFORMATION 
 60   IF (CL) 130,130,70 
 70   IF (NCV.LE.0) RETURN 
      B=UP*U(1,LPIVOT) 
      IF (B) 80,130,130 
 80   B=1.0/B 
      I2=1-ICV+ICE*(LPIVOT-1) 
      INCR=ICE*(L1-LPIVOT) 
      DO 120 J=1,NCV 
      I2=ICV + I2 
      I3=I2 + INCR 
      I4=I3 
      SM=C(I2)*UP 
      DO 90 I=L1,M 
      SM=C(I3)*U(1,I) + SM 
 90   I3=ICE + I3 
      IF (SM) 100,120,100 
 100  SM=B*SM 
      C(I2)=SM*UP + C(I2) 
      DO 110 I=L1,M 
      C(I4)=SM*U(1,I) + C(I4) 
 110  I4=ICE + I4 
 120  CONTINUE 
 130  RETURN 
      END 
C      SUBROUTINE DSIGN(X,Y) 
C	  DOUBLE PRECISION X, Y, Z
C	  Z =  ABS(X)*SIGN(Y)
C	  RETURN Z
C	  END
