import javax.swing.*;
import javax.swing.event.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

class DisplayMIF extends JFrame implements KeyListener, MouseListener, MouseMotionListener {

	private ImagePanel imagePanel;
	private JSlider sliceSlider;
	private ChartPanel chartPanel;
	private int width, height, nslices;
	private int[] imageData;
	private Point mouseLoc;

	//-----------------------------------------------------
	//
	//  Constructor
	//
	DisplayMIF(String title, Volume data) {

		//
		//  Tells the program to exit when this application is closed.
		//
		setDefaultCloseOperation(3);

		width = data.nX();
		height = data.nY();
		nslices = data.nZ();

		//
		//  Copy the data over (scaled).
		//
		int max= data.getMax();

		//
		//  Scale and set the imageData variable.
		//
		imageData = new int[width*height*nslices];
		for(int ii=0; ii<width*height*nslices; ii++) {
			imageData[ii] = (int)((float)data.getElement(ii) / (float)max * 255.0);
		}

		//
		//  Set the true imageData variable to RGB.
		//
		int jj;
		for(int ii=0; ii<width*height*nslices; ii++) {
			jj = imageData[ii] & 0xff;
			imageData[ii] = ((255<<24) | (jj<<16) | (jj<<8) | jj);
		}

		
		//
		//  Set the mouse and key listeners.
		//
		addKeyListener(this);
		addMouseListener(this);
		addMouseMotionListener(this);

		getContentPane().setLayout(new BorderLayout(2,2));

		//
		//  Create the JTextArea
		//
		// textArea = new JTextPane();
		// textArea.setPreferredSize(new Dimension( 200, height ) );
		// getContentPane().add(textArea, BorderLayout.WEST);
		// textArea.setBorder(BorderFactory.createTitledBorder("Info"));
		// textArea.setEditable( false );
		chartPanel = new ChartPanel();
		chartPanel.setPreferredSize(new Dimension( 200, height ) );
		getContentPane().add(chartPanel, BorderLayout.WEST);
		chartPanel.setBorder(BorderFactory.createTitledBorder("Info"));

		//
		//  Create the panel.
		//
		imagePanel = new ImagePanel(imageData, width, height, nslices);
		imagePanel.setSize(width, height);
		imagePanel.setPreferredSize(new Dimension(width, height));
		getContentPane().add(imagePanel, BorderLayout.EAST);
		imagePanel.setBorder(BorderFactory.createTitledBorder("Image"));

		
		//
		//  Create the slider.
		//
		sliceSlider = new JSlider(JSlider.HORIZONTAL, 1, nslices, 1);
		imagePanel.setSize(256,10);
		getContentPane().add(sliceSlider, BorderLayout.SOUTH);
		sliceSlider.setBorder(BorderFactory.createTitledBorder("Slice Number"));

		sliceSlider.setMajorTickSpacing(5);
		sliceSlider.setMinorTickSpacing(1);
		sliceSlider.setPaintTicks(true);
		sliceSlider.setPaintLabels(true);
		sliceSlider.addChangeListener(new SliceListener());

		//
		//  Set the size and display the frame.
		//
//		setSize(260,350);  // was this 
		setSize(260+200,350);
		setTitle(title);
//		setResizable(false);
		pack();
		show();

		int y[] = new int[256];
		for(int ii=0; ii<256; ii++) {
			y[ii] = imageData[ii+256*128];
		}
		chartPanel.setY(y);
	}

	//-----------------------------------------------------
	//
	//  Constructor
	//
	DisplayMIF(String title, int[] id, int w, int h, int slices) {

		//
		//  Tells the program to exit when this application is closed.
		//
		setDefaultCloseOperation(3);

		width = w;
		height = h;
		nslices = slices;

		//
		//  Copy the data over (scaled).
		//
		int max=0;
		for(int ii=0; ii<width*height*nslices; ii++) {
			if( id[ii] > max ) {	
				max = id[ii];
			}
		}
		// System.out.println("Maximum value is " + max);

		//
		//  Scale and set the imageData variable.
		//
		imageData = new int[width*height*nslices];
		for(int ii=0; ii<width*height*nslices; ii++) {
			imageData[ii] = (int)((float)id[ii] / (float)max * 255.0);
		}

		//
		//  Set the true imageData variable to RGB.
		//
		int jj;
		for(int ii=0; ii<width*height*nslices; ii++) {
			jj = imageData[ii] & 0xff;
			imageData[ii] = ((255<<24) | (jj<<16) | (jj<<8) | jj);
		}

		
		//
		//  Set the mouse and key listeners.
		//
		addKeyListener(this);
		addMouseListener(this);
		addMouseMotionListener(this);

		getContentPane().setLayout(new BorderLayout(2,1));

		//
		//  Create the panel.
		//
		imagePanel = new ImagePanel(imageData, width, height, nslices);
		imagePanel.setSize(width, height);
		imagePanel.setPreferredSize(new Dimension(width, height));
		getContentPane().add(imagePanel, BorderLayout.NORTH);
		imagePanel.setBorder(BorderFactory.createTitledBorder("Image"));


		
		//
		//  Create the slider.
		//
		sliceSlider = new JSlider(JSlider.HORIZONTAL, 1, nslices, 1);
		imagePanel.setSize(256,10);
		getContentPane().add(sliceSlider, BorderLayout.SOUTH);
		sliceSlider.setBorder(BorderFactory.createTitledBorder("Slice Number"));

		sliceSlider.setMajorTickSpacing(5);
		sliceSlider.setMinorTickSpacing(1);
		sliceSlider.setPaintTicks(true);
		sliceSlider.setPaintLabels(true);
		sliceSlider.addChangeListener(new SliceListener());

		//
		//  Set the size and display the frame.
		//
		setSize(260,350);
		setTitle(title);
		setResizable(false);
		show();
	}

	protected void paintComponent(Graphics g) {
		imagePanel.paintComponent(g);
		chartPanel.paintComponent(g);
	}

	//-----------------------------------------------------
	//
	//  Mouse Listeners
	//
    public void mouseReleased(MouseEvent e) {}
    public void mousePressed(MouseEvent e) {
        mouseLoc = e.getPoint();
		String temp = "(" + mouseLoc.x + ", " +mouseLoc.y+")";
		//textArea.setText(temp);

		int y[] = new int[256];
		for(int ii=0; ii<256; ii++) {
			y[ii] = imageData[ii+256*mouseLoc.y];
		}
		chartPanel.setY(y);
		chartPanel.repaint();
        System.out.println("(" + mouseLoc.x + ", " +mouseLoc.y+").");
    }

    public void mouseExited(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseClicked(MouseEvent e) {}

	public void mouseDragged(MouseEvent e) {
        mouseLoc = e.getPoint();
		String temp = "(" + mouseLoc.x + ", " +mouseLoc.y+")";
		//textArea.setText(temp);
	}

	public void mouseMoved(MouseEvent e) {
		mouseLoc = e.getPoint();
	}

	public void keyPressed(KeyEvent e) {}
	public void keyReleased(KeyEvent e) {}
	public void keyTyped(KeyEvent e) {
        System.out.println("The " + e.getKeyChar() + " key pressed (" + mouseLoc.x + ", " +mouseLoc.y+").");

		if( e.getKeyChar() == 'q' || e.getKeyChar() == 'Q' ) {
			System.exit(0);
		}
	}

	class SliceListener implements ChangeListener {
		
		public void stateChanged(ChangeEvent e) {
			JSlider source = (JSlider)e.getSource();
			imagePanel.displaySlice(source.getValue());
		}
	}

}

class ImagePanel extends JPanel {

	private MemoryImageSource mi;
	private Image m_image;
	int width, height, nslices;
	int imageData[];

	ImagePanel(int[] p_imageData, int p_width, int p_height, int p_nslices) {

		//
		//  Copy the width and height to internal variables.
		//
		width = p_width;
		height = p_height;
		nslices = p_nslices;

		imageData = new int[width*height*nslices];

		//
		//  Copy the data over.
		//
		for(int ii=0; ii<width*height*nslices;ii++) {
			imageData[ii] = p_imageData[ii];
		}

		mi = new MemoryImageSource(width, height, imageData, 0, width);
		displaySlice(1);
		mi.setAnimated(true);

	}

	public void displaySlice(int sliceNumber) {

		mi.newPixels(imageData, getColorModel(), 
			(sliceNumber-1)*width*height, width);

		m_image = createImage(mi);

	}

	public void paintComponent(Graphics g) {
		g.drawImage(m_image, 0, 0, width, height, this);
	}
}
