package apps.t2part;

import java.io.*;
import java.util.*;
import cj.fileformats.*;
import cj.volume.*;
import gnu.getopt.*;

class T2PartApp {

	static protected void usage()
	{
		System.out.println("t2part [options] <filename>");
		System.out.println("Options:");
		System.out.println("\t[-b|--bin] <name> <start> <end> - Name of bin and starting/ending times.");

		System.out.println("");
		System.out.println("Notes:");
		System.out.println("  - Multiple -b <name> <start> <end> will be needed.");
		System.out.println("  - Should have a -b g 0 99999999999 for the global bin");
	}

	public static void main(String args[]) {
		String filename = "";
        short verbose = 0;
		Vector names = new Vector(10);
		Vector starts = new Vector(10);
		Vector stops = new Vector(10);
 
        //-------------------------------------------------------
        //
        //  Check the arguements to the program.
        //
        LongOpt[] longopts = new LongOpt[3];
        longopts[0] = new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h');
        longopts[1] = new LongOpt("verbose", LongOpt.NO_ARGUMENT, null, 'v');
        longopts[2] = new LongOpt("bin", LongOpt.REQUIRED_ARGUMENT, null, 'b');

        Getopt g = new Getopt("T2part", args, "hvb:", longopts);
        int c;
        while((c=g.getopt()) != -1 ) {
            switch(c) {
                case 'h':
					usage();
                    System.exit(0);
                break;
                case 'v':
                    verbose++;
                break;
                case 'b':
					names.add( g.getOptarg() );

					//  Make sure the start thing is a number.
					try {
						starts.add( new Double( args[g.getOptind()+0] ));
					}
					catch( NumberFormatException e )
					{
						System.out.println("T2Part:  " + args[g.getOptind()+0] + " is not a number.");
						System.exit(-1);
					}

					//  Make sure the stop thing is a number.
					try {
						stops.add( new Double( args[g.getOptind()+1] ) );
					}
					catch( NumberFormatException e )
					{
						System.out.println("T2Part:  " + args[g.getOptind()+1] + " is not a number.");
						System.exit(-1);
					}

					g.setOptind( g.getOptind()+2 );
                break;
            }
        }

        if( g.getOptind() >= args.length ) 
		{
            usage();
			System.exit(0);
        }

        if( g.getOptind() < args.length ) {
            filename = args[g.getOptind()];
        }

		//
		//  Display the information parsed from the command line.
		//
		if( verbose > 0 )
		{
			System.out.println("Filename is " + filename );
			for(int ii=0; ii<names.size(); ii++) {
				System.out.println("Bin: " + (String)names.elementAt(ii) + "(" + 
					((Double)starts.elementAt(ii)).doubleValue() + "," +
					((Double)stops.elementAt(ii)).doubleValue() + "]");
			}
		}

		//
		//  Read in the BFFSpectral file.
		//
		BFFSpectral bffs = new BFFSpectral( filename );

		//
		//  Get the spectral volume.
		//
		SpectralVolume sv = bffs.getSpectralVolume();

		//
		//  Now add the bins.
		//
		for(int ii=0; ii<names.size(); ii++) {
			sv.defineBin(((Double)starts.elementAt(ii)).doubleValue(),
			            ((Double)stops.elementAt(ii)).doubleValue(), 
						(String)names.elementAt(ii));
		}

		//
		//  Now compute the bins.
		//
		sv.computeBins( bffs.getT2() );

		//
		//  Save the bins.
		//
		int begin = filename.lastIndexOf("/");
		int end = filename.indexOf(".bff");
		String basename = filename.substring(begin+1, end);
		sv.saveBins( basename, bffs.getHeader() );
	}
}
